# Generated from MalloyParser.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .MalloyParser import MalloyParser
else:
    from MalloyParser import MalloyParser

# This class defines a complete listener for a parse tree produced by MalloyParser.
class MalloyParserListener(ParseTreeListener):

    # Enter a parse tree produced by MalloyParser#malloyDocument.
    def enterMalloyDocument(self, ctx:MalloyParser.MalloyDocumentContext):
        pass

    # Exit a parse tree produced by MalloyParser#malloyDocument.
    def exitMalloyDocument(self, ctx:MalloyParser.MalloyDocumentContext):
        pass


    # Enter a parse tree produced by MalloyParser#malloyStatement.
    def enterMalloyStatement(self, ctx:MalloyParser.MalloyStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#malloyStatement.
    def exitMalloyStatement(self, ctx:MalloyParser.MalloyStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#defineSourceStatement.
    def enterDefineSourceStatement(self, ctx:MalloyParser.DefineSourceStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#defineSourceStatement.
    def exitDefineSourceStatement(self, ctx:MalloyParser.DefineSourceStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#use_top_level_query_defs.
    def enterUse_top_level_query_defs(self, ctx:MalloyParser.Use_top_level_query_defsContext):
        pass

    # Exit a parse tree produced by MalloyParser#use_top_level_query_defs.
    def exitUse_top_level_query_defs(self, ctx:MalloyParser.Use_top_level_query_defsContext):
        pass


    # Enter a parse tree produced by MalloyParser#topLevelAnonQueryDef.
    def enterTopLevelAnonQueryDef(self, ctx:MalloyParser.TopLevelAnonQueryDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#topLevelAnonQueryDef.
    def exitTopLevelAnonQueryDef(self, ctx:MalloyParser.TopLevelAnonQueryDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#tags.
    def enterTags(self, ctx:MalloyParser.TagsContext):
        pass

    # Exit a parse tree produced by MalloyParser#tags.
    def exitTags(self, ctx:MalloyParser.TagsContext):
        pass


    # Enter a parse tree produced by MalloyParser#isDefine.
    def enterIsDefine(self, ctx:MalloyParser.IsDefineContext):
        pass

    # Exit a parse tree produced by MalloyParser#isDefine.
    def exitIsDefine(self, ctx:MalloyParser.IsDefineContext):
        pass


    # Enter a parse tree produced by MalloyParser#runStatement.
    def enterRunStatement(self, ctx:MalloyParser.RunStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#runStatement.
    def exitRunStatement(self, ctx:MalloyParser.RunStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#sqlString.
    def enterSqlString(self, ctx:MalloyParser.SqlStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#sqlString.
    def exitSqlString(self, ctx:MalloyParser.SqlStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#sqlInterpolation.
    def enterSqlInterpolation(self, ctx:MalloyParser.SqlInterpolationContext):
        pass

    # Exit a parse tree produced by MalloyParser#sqlInterpolation.
    def exitSqlInterpolation(self, ctx:MalloyParser.SqlInterpolationContext):
        pass


    # Enter a parse tree produced by MalloyParser#importStatement.
    def enterImportStatement(self, ctx:MalloyParser.ImportStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#importStatement.
    def exitImportStatement(self, ctx:MalloyParser.ImportStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#importSelect.
    def enterImportSelect(self, ctx:MalloyParser.ImportSelectContext):
        pass

    # Exit a parse tree produced by MalloyParser#importSelect.
    def exitImportSelect(self, ctx:MalloyParser.ImportSelectContext):
        pass


    # Enter a parse tree produced by MalloyParser#importItem.
    def enterImportItem(self, ctx:MalloyParser.ImportItemContext):
        pass

    # Exit a parse tree produced by MalloyParser#importItem.
    def exitImportItem(self, ctx:MalloyParser.ImportItemContext):
        pass


    # Enter a parse tree produced by MalloyParser#importURL.
    def enterImportURL(self, ctx:MalloyParser.ImportURLContext):
        pass

    # Exit a parse tree produced by MalloyParser#importURL.
    def exitImportURL(self, ctx:MalloyParser.ImportURLContext):
        pass


    # Enter a parse tree produced by MalloyParser#docAnnotations.
    def enterDocAnnotations(self, ctx:MalloyParser.DocAnnotationsContext):
        pass

    # Exit a parse tree produced by MalloyParser#docAnnotations.
    def exitDocAnnotations(self, ctx:MalloyParser.DocAnnotationsContext):
        pass


    # Enter a parse tree produced by MalloyParser#ignoredObjectAnnotations.
    def enterIgnoredObjectAnnotations(self, ctx:MalloyParser.IgnoredObjectAnnotationsContext):
        pass

    # Exit a parse tree produced by MalloyParser#ignoredObjectAnnotations.
    def exitIgnoredObjectAnnotations(self, ctx:MalloyParser.IgnoredObjectAnnotationsContext):
        pass


    # Enter a parse tree produced by MalloyParser#ignoredModelAnnotations.
    def enterIgnoredModelAnnotations(self, ctx:MalloyParser.IgnoredModelAnnotationsContext):
        pass

    # Exit a parse tree produced by MalloyParser#ignoredModelAnnotations.
    def exitIgnoredModelAnnotations(self, ctx:MalloyParser.IgnoredModelAnnotationsContext):
        pass


    # Enter a parse tree produced by MalloyParser#topLevelQueryDefs.
    def enterTopLevelQueryDefs(self, ctx:MalloyParser.TopLevelQueryDefsContext):
        pass

    # Exit a parse tree produced by MalloyParser#topLevelQueryDefs.
    def exitTopLevelQueryDefs(self, ctx:MalloyParser.TopLevelQueryDefsContext):
        pass


    # Enter a parse tree produced by MalloyParser#topLevelQueryDef.
    def enterTopLevelQueryDef(self, ctx:MalloyParser.TopLevelQueryDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#topLevelQueryDef.
    def exitTopLevelQueryDef(self, ctx:MalloyParser.TopLevelQueryDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#refineOperator.
    def enterRefineOperator(self, ctx:MalloyParser.RefineOperatorContext):
        pass

    # Exit a parse tree produced by MalloyParser#refineOperator.
    def exitRefineOperator(self, ctx:MalloyParser.RefineOperatorContext):
        pass


    # Enter a parse tree produced by MalloyParser#turtleName.
    def enterTurtleName(self, ctx:MalloyParser.TurtleNameContext):
        pass

    # Exit a parse tree produced by MalloyParser#turtleName.
    def exitTurtleName(self, ctx:MalloyParser.TurtleNameContext):
        pass


    # Enter a parse tree produced by MalloyParser#sqlSource.
    def enterSqlSource(self, ctx:MalloyParser.SqlSourceContext):
        pass

    # Exit a parse tree produced by MalloyParser#sqlSource.
    def exitSqlSource(self, ctx:MalloyParser.SqlSourceContext):
        pass


    # Enter a parse tree produced by MalloyParser#exploreTable.
    def enterExploreTable(self, ctx:MalloyParser.ExploreTableContext):
        pass

    # Exit a parse tree produced by MalloyParser#exploreTable.
    def exitExploreTable(self, ctx:MalloyParser.ExploreTableContext):
        pass


    # Enter a parse tree produced by MalloyParser#connectionId.
    def enterConnectionId(self, ctx:MalloyParser.ConnectionIdContext):
        pass

    # Exit a parse tree produced by MalloyParser#connectionId.
    def exitConnectionId(self, ctx:MalloyParser.ConnectionIdContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryProperties.
    def enterQueryProperties(self, ctx:MalloyParser.QueryPropertiesContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryProperties.
    def exitQueryProperties(self, ctx:MalloyParser.QueryPropertiesContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryName.
    def enterQueryName(self, ctx:MalloyParser.QueryNameContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryName.
    def exitQueryName(self, ctx:MalloyParser.QueryNameContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourcePropertyList.
    def enterSourcePropertyList(self, ctx:MalloyParser.SourcePropertyListContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourcePropertyList.
    def exitSourcePropertyList(self, ctx:MalloyParser.SourcePropertyListContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceDefinition.
    def enterSourceDefinition(self, ctx:MalloyParser.SourceDefinitionContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceDefinition.
    def exitSourceDefinition(self, ctx:MalloyParser.SourceDefinitionContext):
        pass


    # Enter a parse tree produced by MalloyParser#sqExplore.
    def enterSqExplore(self, ctx:MalloyParser.SqExploreContext):
        pass

    # Exit a parse tree produced by MalloyParser#sqExplore.
    def exitSqExplore(self, ctx:MalloyParser.SqExploreContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceParameters.
    def enterSourceParameters(self, ctx:MalloyParser.SourceParametersContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceParameters.
    def exitSourceParameters(self, ctx:MalloyParser.SourceParametersContext):
        pass


    # Enter a parse tree produced by MalloyParser#legalParamType.
    def enterLegalParamType(self, ctx:MalloyParser.LegalParamTypeContext):
        pass

    # Exit a parse tree produced by MalloyParser#legalParamType.
    def exitLegalParamType(self, ctx:MalloyParser.LegalParamTypeContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceParameter.
    def enterSourceParameter(self, ctx:MalloyParser.SourceParameterContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceParameter.
    def exitSourceParameter(self, ctx:MalloyParser.SourceParameterContext):
        pass


    # Enter a parse tree produced by MalloyParser#parameterNameDef.
    def enterParameterNameDef(self, ctx:MalloyParser.ParameterNameDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#parameterNameDef.
    def exitParameterNameDef(self, ctx:MalloyParser.ParameterNameDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceNameDef.
    def enterSourceNameDef(self, ctx:MalloyParser.SourceNameDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceNameDef.
    def exitSourceNameDef(self, ctx:MalloyParser.SourceNameDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#exploreProperties.
    def enterExploreProperties(self, ctx:MalloyParser.ExplorePropertiesContext):
        pass

    # Exit a parse tree produced by MalloyParser#exploreProperties.
    def exitExploreProperties(self, ctx:MalloyParser.ExplorePropertiesContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreDimension_stub.
    def enterDefExploreDimension_stub(self, ctx:MalloyParser.DefExploreDimension_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreDimension_stub.
    def exitDefExploreDimension_stub(self, ctx:MalloyParser.DefExploreDimension_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreMeasure_stub.
    def enterDefExploreMeasure_stub(self, ctx:MalloyParser.DefExploreMeasure_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreMeasure_stub.
    def exitDefExploreMeasure_stub(self, ctx:MalloyParser.DefExploreMeasure_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#defJoin_stub.
    def enterDefJoin_stub(self, ctx:MalloyParser.DefJoin_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#defJoin_stub.
    def exitDefJoin_stub(self, ctx:MalloyParser.DefJoin_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreWhere_stub.
    def enterDefExploreWhere_stub(self, ctx:MalloyParser.DefExploreWhere_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreWhere_stub.
    def exitDefExploreWhere_stub(self, ctx:MalloyParser.DefExploreWhere_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExplorePrimaryKey.
    def enterDefExplorePrimaryKey(self, ctx:MalloyParser.DefExplorePrimaryKeyContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExplorePrimaryKey.
    def exitDefExplorePrimaryKey(self, ctx:MalloyParser.DefExplorePrimaryKeyContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreRename.
    def enterDefExploreRename(self, ctx:MalloyParser.DefExploreRenameContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreRename.
    def exitDefExploreRename(self, ctx:MalloyParser.DefExploreRenameContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreEditField.
    def enterDefExploreEditField(self, ctx:MalloyParser.DefExploreEditFieldContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreEditField.
    def exitDefExploreEditField(self, ctx:MalloyParser.DefExploreEditFieldContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreQuery.
    def enterDefExploreQuery(self, ctx:MalloyParser.DefExploreQueryContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreQuery.
    def exitDefExploreQuery(self, ctx:MalloyParser.DefExploreQueryContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreTimezone.
    def enterDefExploreTimezone(self, ctx:MalloyParser.DefExploreTimezoneContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreTimezone.
    def exitDefExploreTimezone(self, ctx:MalloyParser.DefExploreTimezoneContext):
        pass


    # Enter a parse tree produced by MalloyParser#defExploreAnnotation.
    def enterDefExploreAnnotation(self, ctx:MalloyParser.DefExploreAnnotationContext):
        pass

    # Exit a parse tree produced by MalloyParser#defExploreAnnotation.
    def exitDefExploreAnnotation(self, ctx:MalloyParser.DefExploreAnnotationContext):
        pass


    # Enter a parse tree produced by MalloyParser#defIgnoreModel_stub.
    def enterDefIgnoreModel_stub(self, ctx:MalloyParser.DefIgnoreModel_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#defIgnoreModel_stub.
    def exitDefIgnoreModel_stub(self, ctx:MalloyParser.DefIgnoreModel_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#accessLabel.
    def enterAccessLabel(self, ctx:MalloyParser.AccessLabelContext):
        pass

    # Exit a parse tree produced by MalloyParser#accessLabel.
    def exitAccessLabel(self, ctx:MalloyParser.AccessLabelContext):
        pass


    # Enter a parse tree produced by MalloyParser#accessModifierList.
    def enterAccessModifierList(self, ctx:MalloyParser.AccessModifierListContext):
        pass

    # Exit a parse tree produced by MalloyParser#accessModifierList.
    def exitAccessModifierList(self, ctx:MalloyParser.AccessModifierListContext):
        pass


    # Enter a parse tree produced by MalloyParser#defMeasures.
    def enterDefMeasures(self, ctx:MalloyParser.DefMeasuresContext):
        pass

    # Exit a parse tree produced by MalloyParser#defMeasures.
    def exitDefMeasures(self, ctx:MalloyParser.DefMeasuresContext):
        pass


    # Enter a parse tree produced by MalloyParser#defDimensions.
    def enterDefDimensions(self, ctx:MalloyParser.DefDimensionsContext):
        pass

    # Exit a parse tree produced by MalloyParser#defDimensions.
    def exitDefDimensions(self, ctx:MalloyParser.DefDimensionsContext):
        pass


    # Enter a parse tree produced by MalloyParser#renameList.
    def enterRenameList(self, ctx:MalloyParser.RenameListContext):
        pass

    # Exit a parse tree produced by MalloyParser#renameList.
    def exitRenameList(self, ctx:MalloyParser.RenameListContext):
        pass


    # Enter a parse tree produced by MalloyParser#renameEntry.
    def enterRenameEntry(self, ctx:MalloyParser.RenameEntryContext):
        pass

    # Exit a parse tree produced by MalloyParser#renameEntry.
    def exitRenameEntry(self, ctx:MalloyParser.RenameEntryContext):
        pass


    # Enter a parse tree produced by MalloyParser#defList.
    def enterDefList(self, ctx:MalloyParser.DefListContext):
        pass

    # Exit a parse tree produced by MalloyParser#defList.
    def exitDefList(self, ctx:MalloyParser.DefListContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldDef.
    def enterFieldDef(self, ctx:MalloyParser.FieldDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldDef.
    def exitFieldDef(self, ctx:MalloyParser.FieldDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldNameDef.
    def enterFieldNameDef(self, ctx:MalloyParser.FieldNameDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldNameDef.
    def exitFieldNameDef(self, ctx:MalloyParser.FieldNameDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinNameDef.
    def enterJoinNameDef(self, ctx:MalloyParser.JoinNameDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinNameDef.
    def exitJoinNameDef(self, ctx:MalloyParser.JoinNameDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#declareStatement.
    def enterDeclareStatement(self, ctx:MalloyParser.DeclareStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#declareStatement.
    def exitDeclareStatement(self, ctx:MalloyParser.DeclareStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#defJoinOne.
    def enterDefJoinOne(self, ctx:MalloyParser.DefJoinOneContext):
        pass

    # Exit a parse tree produced by MalloyParser#defJoinOne.
    def exitDefJoinOne(self, ctx:MalloyParser.DefJoinOneContext):
        pass


    # Enter a parse tree produced by MalloyParser#defJoinMany.
    def enterDefJoinMany(self, ctx:MalloyParser.DefJoinManyContext):
        pass

    # Exit a parse tree produced by MalloyParser#defJoinMany.
    def exitDefJoinMany(self, ctx:MalloyParser.DefJoinManyContext):
        pass


    # Enter a parse tree produced by MalloyParser#defJoinCross.
    def enterDefJoinCross(self, ctx:MalloyParser.DefJoinCrossContext):
        pass

    # Exit a parse tree produced by MalloyParser#defJoinCross.
    def exitDefJoinCross(self, ctx:MalloyParser.DefJoinCrossContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryExtend.
    def enterQueryExtend(self, ctx:MalloyParser.QueryExtendContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryExtend.
    def exitQueryExtend(self, ctx:MalloyParser.QueryExtendContext):
        pass


    # Enter a parse tree produced by MalloyParser#modEither.
    def enterModEither(self, ctx:MalloyParser.ModEitherContext):
        pass

    # Exit a parse tree produced by MalloyParser#modEither.
    def exitModEither(self, ctx:MalloyParser.ModEitherContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceArguments.
    def enterSourceArguments(self, ctx:MalloyParser.SourceArgumentsContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceArguments.
    def exitSourceArguments(self, ctx:MalloyParser.SourceArgumentsContext):
        pass


    # Enter a parse tree produced by MalloyParser#argumentId.
    def enterArgumentId(self, ctx:MalloyParser.ArgumentIdContext):
        pass

    # Exit a parse tree produced by MalloyParser#argumentId.
    def exitArgumentId(self, ctx:MalloyParser.ArgumentIdContext):
        pass


    # Enter a parse tree produced by MalloyParser#sourceArgument.
    def enterSourceArgument(self, ctx:MalloyParser.SourceArgumentContext):
        pass

    # Exit a parse tree produced by MalloyParser#sourceArgument.
    def exitSourceArgument(self, ctx:MalloyParser.SourceArgumentContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQRefinedQuery.
    def enterSQRefinedQuery(self, ctx:MalloyParser.SQRefinedQueryContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQRefinedQuery.
    def exitSQRefinedQuery(self, ctx:MalloyParser.SQRefinedQueryContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQSQL.
    def enterSQSQL(self, ctx:MalloyParser.SQSQLContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQSQL.
    def exitSQSQL(self, ctx:MalloyParser.SQSQLContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQExtendedSource.
    def enterSQExtendedSource(self, ctx:MalloyParser.SQExtendedSourceContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQExtendedSource.
    def exitSQExtendedSource(self, ctx:MalloyParser.SQExtendedSourceContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQCompose.
    def enterSQCompose(self, ctx:MalloyParser.SQComposeContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQCompose.
    def exitSQCompose(self, ctx:MalloyParser.SQComposeContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQTable.
    def enterSQTable(self, ctx:MalloyParser.SQTableContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQTable.
    def exitSQTable(self, ctx:MalloyParser.SQTableContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQInclude.
    def enterSQInclude(self, ctx:MalloyParser.SQIncludeContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQInclude.
    def exitSQInclude(self, ctx:MalloyParser.SQIncludeContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQArrow.
    def enterSQArrow(self, ctx:MalloyParser.SQArrowContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQArrow.
    def exitSQArrow(self, ctx:MalloyParser.SQArrowContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQParens.
    def enterSQParens(self, ctx:MalloyParser.SQParensContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQParens.
    def exitSQParens(self, ctx:MalloyParser.SQParensContext):
        pass


    # Enter a parse tree produced by MalloyParser#SQID.
    def enterSQID(self, ctx:MalloyParser.SQIDContext):
        pass

    # Exit a parse tree produced by MalloyParser#SQID.
    def exitSQID(self, ctx:MalloyParser.SQIDContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeBlock.
    def enterIncludeBlock(self, ctx:MalloyParser.IncludeBlockContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeBlock.
    def exitIncludeBlock(self, ctx:MalloyParser.IncludeBlockContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeItem.
    def enterIncludeItem(self, ctx:MalloyParser.IncludeItemContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeItem.
    def exitIncludeItem(self, ctx:MalloyParser.IncludeItemContext):
        pass


    # Enter a parse tree produced by MalloyParser#orphanedAnnotation.
    def enterOrphanedAnnotation(self, ctx:MalloyParser.OrphanedAnnotationContext):
        pass

    # Exit a parse tree produced by MalloyParser#orphanedAnnotation.
    def exitOrphanedAnnotation(self, ctx:MalloyParser.OrphanedAnnotationContext):
        pass


    # Enter a parse tree produced by MalloyParser#accessLabelProp.
    def enterAccessLabelProp(self, ctx:MalloyParser.AccessLabelPropContext):
        pass

    # Exit a parse tree produced by MalloyParser#accessLabelProp.
    def exitAccessLabelProp(self, ctx:MalloyParser.AccessLabelPropContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeExceptList.
    def enterIncludeExceptList(self, ctx:MalloyParser.IncludeExceptListContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeExceptList.
    def exitIncludeExceptList(self, ctx:MalloyParser.IncludeExceptListContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeExceptListItem.
    def enterIncludeExceptListItem(self, ctx:MalloyParser.IncludeExceptListItemContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeExceptListItem.
    def exitIncludeExceptListItem(self, ctx:MalloyParser.IncludeExceptListItemContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeList.
    def enterIncludeList(self, ctx:MalloyParser.IncludeListContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeList.
    def exitIncludeList(self, ctx:MalloyParser.IncludeListContext):
        pass


    # Enter a parse tree produced by MalloyParser#includeField.
    def enterIncludeField(self, ctx:MalloyParser.IncludeFieldContext):
        pass

    # Exit a parse tree produced by MalloyParser#includeField.
    def exitIncludeField(self, ctx:MalloyParser.IncludeFieldContext):
        pass


    # Enter a parse tree produced by MalloyParser#SegField.
    def enterSegField(self, ctx:MalloyParser.SegFieldContext):
        pass

    # Exit a parse tree produced by MalloyParser#SegField.
    def exitSegField(self, ctx:MalloyParser.SegFieldContext):
        pass


    # Enter a parse tree produced by MalloyParser#SegRefine.
    def enterSegRefine(self, ctx:MalloyParser.SegRefineContext):
        pass

    # Exit a parse tree produced by MalloyParser#SegRefine.
    def exitSegRefine(self, ctx:MalloyParser.SegRefineContext):
        pass


    # Enter a parse tree produced by MalloyParser#SegParen.
    def enterSegParen(self, ctx:MalloyParser.SegParenContext):
        pass

    # Exit a parse tree produced by MalloyParser#SegParen.
    def exitSegParen(self, ctx:MalloyParser.SegParenContext):
        pass


    # Enter a parse tree produced by MalloyParser#SegOps.
    def enterSegOps(self, ctx:MalloyParser.SegOpsContext):
        pass

    # Exit a parse tree produced by MalloyParser#SegOps.
    def exitSegOps(self, ctx:MalloyParser.SegOpsContext):
        pass


    # Enter a parse tree produced by MalloyParser#VSeg.
    def enterVSeg(self, ctx:MalloyParser.VSegContext):
        pass

    # Exit a parse tree produced by MalloyParser#VSeg.
    def exitVSeg(self, ctx:MalloyParser.VSegContext):
        pass


    # Enter a parse tree produced by MalloyParser#VArrow.
    def enterVArrow(self, ctx:MalloyParser.VArrowContext):
        pass

    # Exit a parse tree produced by MalloyParser#VArrow.
    def exitVArrow(self, ctx:MalloyParser.VArrowContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryExtendStatement.
    def enterQueryExtendStatement(self, ctx:MalloyParser.QueryExtendStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryExtendStatement.
    def exitQueryExtendStatement(self, ctx:MalloyParser.QueryExtendStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryExtendStatementList.
    def enterQueryExtendStatementList(self, ctx:MalloyParser.QueryExtendStatementListContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryExtendStatementList.
    def exitQueryExtendStatementList(self, ctx:MalloyParser.QueryExtendStatementListContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinList.
    def enterJoinList(self, ctx:MalloyParser.JoinListContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinList.
    def exitJoinList(self, ctx:MalloyParser.JoinListContext):
        pass


    # Enter a parse tree produced by MalloyParser#isExplore.
    def enterIsExplore(self, ctx:MalloyParser.IsExploreContext):
        pass

    # Exit a parse tree produced by MalloyParser#isExplore.
    def exitIsExplore(self, ctx:MalloyParser.IsExploreContext):
        pass


    # Enter a parse tree produced by MalloyParser#matrixOperation.
    def enterMatrixOperation(self, ctx:MalloyParser.MatrixOperationContext):
        pass

    # Exit a parse tree produced by MalloyParser#matrixOperation.
    def exitMatrixOperation(self, ctx:MalloyParser.MatrixOperationContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinFrom.
    def enterJoinFrom(self, ctx:MalloyParser.JoinFromContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinFrom.
    def exitJoinFrom(self, ctx:MalloyParser.JoinFromContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinWith.
    def enterJoinWith(self, ctx:MalloyParser.JoinWithContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinWith.
    def exitJoinWith(self, ctx:MalloyParser.JoinWithContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinOn.
    def enterJoinOn(self, ctx:MalloyParser.JoinOnContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinOn.
    def exitJoinOn(self, ctx:MalloyParser.JoinOnContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinExpression.
    def enterJoinExpression(self, ctx:MalloyParser.JoinExpressionContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinExpression.
    def exitJoinExpression(self, ctx:MalloyParser.JoinExpressionContext):
        pass


    # Enter a parse tree produced by MalloyParser#filterStatement.
    def enterFilterStatement(self, ctx:MalloyParser.FilterStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#filterStatement.
    def exitFilterStatement(self, ctx:MalloyParser.FilterStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldProperties.
    def enterFieldProperties(self, ctx:MalloyParser.FieldPropertiesContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldProperties.
    def exitFieldProperties(self, ctx:MalloyParser.FieldPropertiesContext):
        pass


    # Enter a parse tree produced by MalloyParser#aggregateOrdering.
    def enterAggregateOrdering(self, ctx:MalloyParser.AggregateOrderingContext):
        pass

    # Exit a parse tree produced by MalloyParser#aggregateOrdering.
    def exitAggregateOrdering(self, ctx:MalloyParser.AggregateOrderingContext):
        pass


    # Enter a parse tree produced by MalloyParser#aggregateOrderBySpec.
    def enterAggregateOrderBySpec(self, ctx:MalloyParser.AggregateOrderBySpecContext):
        pass

    # Exit a parse tree produced by MalloyParser#aggregateOrderBySpec.
    def exitAggregateOrderBySpec(self, ctx:MalloyParser.AggregateOrderBySpecContext):
        pass


    # Enter a parse tree produced by MalloyParser#aggregateOrderByStatement.
    def enterAggregateOrderByStatement(self, ctx:MalloyParser.AggregateOrderByStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#aggregateOrderByStatement.
    def exitAggregateOrderByStatement(self, ctx:MalloyParser.AggregateOrderByStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldPropertyLimitStatement.
    def enterFieldPropertyLimitStatement(self, ctx:MalloyParser.FieldPropertyLimitStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldPropertyLimitStatement.
    def exitFieldPropertyLimitStatement(self, ctx:MalloyParser.FieldPropertyLimitStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldPropertyStatement.
    def enterFieldPropertyStatement(self, ctx:MalloyParser.FieldPropertyStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldPropertyStatement.
    def exitFieldPropertyStatement(self, ctx:MalloyParser.FieldPropertyStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#filterClauseList.
    def enterFilterClauseList(self, ctx:MalloyParser.FilterClauseListContext):
        pass

    # Exit a parse tree produced by MalloyParser#filterClauseList.
    def exitFilterClauseList(self, ctx:MalloyParser.FilterClauseListContext):
        pass


    # Enter a parse tree produced by MalloyParser#whereStatement.
    def enterWhereStatement(self, ctx:MalloyParser.WhereStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#whereStatement.
    def exitWhereStatement(self, ctx:MalloyParser.WhereStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#havingStatement.
    def enterHavingStatement(self, ctx:MalloyParser.HavingStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#havingStatement.
    def exitHavingStatement(self, ctx:MalloyParser.HavingStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#subQueryDefList.
    def enterSubQueryDefList(self, ctx:MalloyParser.SubQueryDefListContext):
        pass

    # Exit a parse tree produced by MalloyParser#subQueryDefList.
    def exitSubQueryDefList(self, ctx:MalloyParser.SubQueryDefListContext):
        pass


    # Enter a parse tree produced by MalloyParser#exploreQueryNameDef.
    def enterExploreQueryNameDef(self, ctx:MalloyParser.ExploreQueryNameDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#exploreQueryNameDef.
    def exitExploreQueryNameDef(self, ctx:MalloyParser.ExploreQueryNameDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#exploreQueryDef.
    def enterExploreQueryDef(self, ctx:MalloyParser.ExploreQueryDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#exploreQueryDef.
    def exitExploreQueryDef(self, ctx:MalloyParser.ExploreQueryDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#drillStatement.
    def enterDrillStatement(self, ctx:MalloyParser.DrillStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#drillStatement.
    def exitDrillStatement(self, ctx:MalloyParser.DrillStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#drillClauseList.
    def enterDrillClauseList(self, ctx:MalloyParser.DrillClauseListContext):
        pass

    # Exit a parse tree produced by MalloyParser#drillClauseList.
    def exitDrillClauseList(self, ctx:MalloyParser.DrillClauseListContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryStatement.
    def enterQueryStatement(self, ctx:MalloyParser.QueryStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryStatement.
    def exitQueryStatement(self, ctx:MalloyParser.QueryStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryJoinStatement.
    def enterQueryJoinStatement(self, ctx:MalloyParser.QueryJoinStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryJoinStatement.
    def exitQueryJoinStatement(self, ctx:MalloyParser.QueryJoinStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#groupByStatement.
    def enterGroupByStatement(self, ctx:MalloyParser.GroupByStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#groupByStatement.
    def exitGroupByStatement(self, ctx:MalloyParser.GroupByStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryFieldList.
    def enterQueryFieldList(self, ctx:MalloyParser.QueryFieldListContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryFieldList.
    def exitQueryFieldList(self, ctx:MalloyParser.QueryFieldListContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryFieldEntry.
    def enterQueryFieldEntry(self, ctx:MalloyParser.QueryFieldEntryContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryFieldEntry.
    def exitQueryFieldEntry(self, ctx:MalloyParser.QueryFieldEntryContext):
        pass


    # Enter a parse tree produced by MalloyParser#nestStatement.
    def enterNestStatement(self, ctx:MalloyParser.NestStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#nestStatement.
    def exitNestStatement(self, ctx:MalloyParser.NestStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#nestedQueryList.
    def enterNestedQueryList(self, ctx:MalloyParser.NestedQueryListContext):
        pass

    # Exit a parse tree produced by MalloyParser#nestedQueryList.
    def exitNestedQueryList(self, ctx:MalloyParser.NestedQueryListContext):
        pass


    # Enter a parse tree produced by MalloyParser#nestDef.
    def enterNestDef(self, ctx:MalloyParser.NestDefContext):
        pass

    # Exit a parse tree produced by MalloyParser#nestDef.
    def exitNestDef(self, ctx:MalloyParser.NestDefContext):
        pass


    # Enter a parse tree produced by MalloyParser#aggregateStatement.
    def enterAggregateStatement(self, ctx:MalloyParser.AggregateStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#aggregateStatement.
    def exitAggregateStatement(self, ctx:MalloyParser.AggregateStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#calculateStatement.
    def enterCalculateStatement(self, ctx:MalloyParser.CalculateStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#calculateStatement.
    def exitCalculateStatement(self, ctx:MalloyParser.CalculateStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#projectStatement.
    def enterProjectStatement(self, ctx:MalloyParser.ProjectStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#projectStatement.
    def exitProjectStatement(self, ctx:MalloyParser.ProjectStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#partitionByStatement.
    def enterPartitionByStatement(self, ctx:MalloyParser.PartitionByStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#partitionByStatement.
    def exitPartitionByStatement(self, ctx:MalloyParser.PartitionByStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#groupedByStatement.
    def enterGroupedByStatement(self, ctx:MalloyParser.GroupedByStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#groupedByStatement.
    def exitGroupedByStatement(self, ctx:MalloyParser.GroupedByStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#orderByStatement.
    def enterOrderByStatement(self, ctx:MalloyParser.OrderByStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#orderByStatement.
    def exitOrderByStatement(self, ctx:MalloyParser.OrderByStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#ordering.
    def enterOrdering(self, ctx:MalloyParser.OrderingContext):
        pass

    # Exit a parse tree produced by MalloyParser#ordering.
    def exitOrdering(self, ctx:MalloyParser.OrderingContext):
        pass


    # Enter a parse tree produced by MalloyParser#orderBySpec.
    def enterOrderBySpec(self, ctx:MalloyParser.OrderBySpecContext):
        pass

    # Exit a parse tree produced by MalloyParser#orderBySpec.
    def exitOrderBySpec(self, ctx:MalloyParser.OrderBySpecContext):
        pass


    # Enter a parse tree produced by MalloyParser#limitStatement.
    def enterLimitStatement(self, ctx:MalloyParser.LimitStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#limitStatement.
    def exitLimitStatement(self, ctx:MalloyParser.LimitStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#bySpec.
    def enterBySpec(self, ctx:MalloyParser.BySpecContext):
        pass

    # Exit a parse tree produced by MalloyParser#bySpec.
    def exitBySpec(self, ctx:MalloyParser.BySpecContext):
        pass


    # Enter a parse tree produced by MalloyParser#topStatement.
    def enterTopStatement(self, ctx:MalloyParser.TopStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#topStatement.
    def exitTopStatement(self, ctx:MalloyParser.TopStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#indexElement.
    def enterIndexElement(self, ctx:MalloyParser.IndexElementContext):
        pass

    # Exit a parse tree produced by MalloyParser#indexElement.
    def exitIndexElement(self, ctx:MalloyParser.IndexElementContext):
        pass


    # Enter a parse tree produced by MalloyParser#indexFields.
    def enterIndexFields(self, ctx:MalloyParser.IndexFieldsContext):
        pass

    # Exit a parse tree produced by MalloyParser#indexFields.
    def exitIndexFields(self, ctx:MalloyParser.IndexFieldsContext):
        pass


    # Enter a parse tree produced by MalloyParser#indexStatement.
    def enterIndexStatement(self, ctx:MalloyParser.IndexStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#indexStatement.
    def exitIndexStatement(self, ctx:MalloyParser.IndexStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#sampleStatement.
    def enterSampleStatement(self, ctx:MalloyParser.SampleStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#sampleStatement.
    def exitSampleStatement(self, ctx:MalloyParser.SampleStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#timezoneStatement.
    def enterTimezoneStatement(self, ctx:MalloyParser.TimezoneStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#timezoneStatement.
    def exitTimezoneStatement(self, ctx:MalloyParser.TimezoneStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#queryAnnotation.
    def enterQueryAnnotation(self, ctx:MalloyParser.QueryAnnotationContext):
        pass

    # Exit a parse tree produced by MalloyParser#queryAnnotation.
    def exitQueryAnnotation(self, ctx:MalloyParser.QueryAnnotationContext):
        pass


    # Enter a parse tree produced by MalloyParser#sampleSpec.
    def enterSampleSpec(self, ctx:MalloyParser.SampleSpecContext):
        pass

    # Exit a parse tree produced by MalloyParser#sampleSpec.
    def exitSampleSpec(self, ctx:MalloyParser.SampleSpecContext):
        pass


    # Enter a parse tree produced by MalloyParser#aggregate.
    def enterAggregate(self, ctx:MalloyParser.AggregateContext):
        pass

    # Exit a parse tree produced by MalloyParser#aggregate.
    def exitAggregate(self, ctx:MalloyParser.AggregateContext):
        pass


    # Enter a parse tree produced by MalloyParser#malloyType.
    def enterMalloyType(self, ctx:MalloyParser.MalloyTypeContext):
        pass

    # Exit a parse tree produced by MalloyParser#malloyType.
    def exitMalloyType(self, ctx:MalloyParser.MalloyTypeContext):
        pass


    # Enter a parse tree produced by MalloyParser#compareOp.
    def enterCompareOp(self, ctx:MalloyParser.CompareOpContext):
        pass

    # Exit a parse tree produced by MalloyParser#compareOp.
    def exitCompareOp(self, ctx:MalloyParser.CompareOpContext):
        pass


    # Enter a parse tree produced by MalloyParser#string.
    def enterString(self, ctx:MalloyParser.StringContext):
        pass

    # Exit a parse tree produced by MalloyParser#string.
    def exitString(self, ctx:MalloyParser.StringContext):
        pass


    # Enter a parse tree produced by MalloyParser#shortString.
    def enterShortString(self, ctx:MalloyParser.ShortStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#shortString.
    def exitShortString(self, ctx:MalloyParser.ShortStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#rawString.
    def enterRawString(self, ctx:MalloyParser.RawStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#rawString.
    def exitRawString(self, ctx:MalloyParser.RawStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#numericLiteral.
    def enterNumericLiteral(self, ctx:MalloyParser.NumericLiteralContext):
        pass

    # Exit a parse tree produced by MalloyParser#numericLiteral.
    def exitNumericLiteral(self, ctx:MalloyParser.NumericLiteralContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprString.
    def enterExprString(self, ctx:MalloyParser.ExprStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprString.
    def exitExprString(self, ctx:MalloyParser.ExprStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#stub_rawString.
    def enterStub_rawString(self, ctx:MalloyParser.Stub_rawStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#stub_rawString.
    def exitStub_rawString(self, ctx:MalloyParser.Stub_rawStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprNumber.
    def enterExprNumber(self, ctx:MalloyParser.ExprNumberContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprNumber.
    def exitExprNumber(self, ctx:MalloyParser.ExprNumberContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprTime.
    def enterExprTime(self, ctx:MalloyParser.ExprTimeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprTime.
    def exitExprTime(self, ctx:MalloyParser.ExprTimeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprNULL.
    def enterExprNULL(self, ctx:MalloyParser.ExprNULLContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprNULL.
    def exitExprNULL(self, ctx:MalloyParser.ExprNULLContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprBool.
    def enterExprBool(self, ctx:MalloyParser.ExprBoolContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprBool.
    def exitExprBool(self, ctx:MalloyParser.ExprBoolContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprRegex.
    def enterExprRegex(self, ctx:MalloyParser.ExprRegexContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprRegex.
    def exitExprRegex(self, ctx:MalloyParser.ExprRegexContext):
        pass


    # Enter a parse tree produced by MalloyParser#filterString_stub.
    def enterFilterString_stub(self, ctx:MalloyParser.FilterString_stubContext):
        pass

    # Exit a parse tree produced by MalloyParser#filterString_stub.
    def exitFilterString_stub(self, ctx:MalloyParser.FilterString_stubContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprNow.
    def enterExprNow(self, ctx:MalloyParser.ExprNowContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprNow.
    def exitExprNow(self, ctx:MalloyParser.ExprNowContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalTimestamp.
    def enterLiteralTimestamp(self, ctx:MalloyParser.LiteralTimestampContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalTimestamp.
    def exitLiteralTimestamp(self, ctx:MalloyParser.LiteralTimestampContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalHour.
    def enterLiteralHour(self, ctx:MalloyParser.LiteralHourContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalHour.
    def exitLiteralHour(self, ctx:MalloyParser.LiteralHourContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalDay.
    def enterLiteralDay(self, ctx:MalloyParser.LiteralDayContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalDay.
    def exitLiteralDay(self, ctx:MalloyParser.LiteralDayContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalWeek.
    def enterLiteralWeek(self, ctx:MalloyParser.LiteralWeekContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalWeek.
    def exitLiteralWeek(self, ctx:MalloyParser.LiteralWeekContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalMonth.
    def enterLiteralMonth(self, ctx:MalloyParser.LiteralMonthContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalMonth.
    def exitLiteralMonth(self, ctx:MalloyParser.LiteralMonthContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalQuarter.
    def enterLiteralQuarter(self, ctx:MalloyParser.LiteralQuarterContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalQuarter.
    def exitLiteralQuarter(self, ctx:MalloyParser.LiteralQuarterContext):
        pass


    # Enter a parse tree produced by MalloyParser#literalYear.
    def enterLiteralYear(self, ctx:MalloyParser.LiteralYearContext):
        pass

    # Exit a parse tree produced by MalloyParser#literalYear.
    def exitLiteralYear(self, ctx:MalloyParser.LiteralYearContext):
        pass


    # Enter a parse tree produced by MalloyParser#tablePath.
    def enterTablePath(self, ctx:MalloyParser.TablePathContext):
        pass

    # Exit a parse tree produced by MalloyParser#tablePath.
    def exitTablePath(self, ctx:MalloyParser.TablePathContext):
        pass


    # Enter a parse tree produced by MalloyParser#tableURI.
    def enterTableURI(self, ctx:MalloyParser.TableURIContext):
        pass

    # Exit a parse tree produced by MalloyParser#tableURI.
    def exitTableURI(self, ctx:MalloyParser.TableURIContext):
        pass


    # Enter a parse tree produced by MalloyParser#id.
    def enterId(self, ctx:MalloyParser.IdContext):
        pass

    # Exit a parse tree produced by MalloyParser#id.
    def exitId(self, ctx:MalloyParser.IdContext):
        pass


    # Enter a parse tree produced by MalloyParser#timeframe.
    def enterTimeframe(self, ctx:MalloyParser.TimeframeContext):
        pass

    # Exit a parse tree produced by MalloyParser#timeframe.
    def exitTimeframe(self, ctx:MalloyParser.TimeframeContext):
        pass


    # Enter a parse tree produced by MalloyParser#ungroup.
    def enterUngroup(self, ctx:MalloyParser.UngroupContext):
        pass

    # Exit a parse tree produced by MalloyParser#ungroup.
    def exitUngroup(self, ctx:MalloyParser.UngroupContext):
        pass


    # Enter a parse tree produced by MalloyParser#malloyOrSQLType.
    def enterMalloyOrSQLType(self, ctx:MalloyParser.MalloyOrSQLTypeContext):
        pass

    # Exit a parse tree produced by MalloyParser#malloyOrSQLType.
    def exitMalloyOrSQLType(self, ctx:MalloyParser.MalloyOrSQLTypeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprExpr.
    def enterExprExpr(self, ctx:MalloyParser.ExprExprContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprExpr.
    def exitExprExpr(self, ctx:MalloyParser.ExprExprContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprCase.
    def enterExprCase(self, ctx:MalloyParser.ExprCaseContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprCase.
    def exitExprCase(self, ctx:MalloyParser.ExprCaseContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprMinus.
    def enterExprMinus(self, ctx:MalloyParser.ExprMinusContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprMinus.
    def exitExprMinus(self, ctx:MalloyParser.ExprMinusContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprAddSub.
    def enterExprAddSub(self, ctx:MalloyParser.ExprAddSubContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprAddSub.
    def exitExprAddSub(self, ctx:MalloyParser.ExprAddSubContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprNullCheck.
    def enterExprNullCheck(self, ctx:MalloyParser.ExprNullCheckContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprNullCheck.
    def exitExprNullCheck(self, ctx:MalloyParser.ExprNullCheckContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprArrayLiteral.
    def enterExprArrayLiteral(self, ctx:MalloyParser.ExprArrayLiteralContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprArrayLiteral.
    def exitExprArrayLiteral(self, ctx:MalloyParser.ExprArrayLiteralContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprRange.
    def enterExprRange(self, ctx:MalloyParser.ExprRangeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprRange.
    def exitExprRange(self, ctx:MalloyParser.ExprRangeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprLogicalOr.
    def enterExprLogicalOr(self, ctx:MalloyParser.ExprLogicalOrContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprLogicalOr.
    def exitExprLogicalOr(self, ctx:MalloyParser.ExprLogicalOrContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprLiteral.
    def enterExprLiteral(self, ctx:MalloyParser.ExprLiteralContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprLiteral.
    def exitExprLiteral(self, ctx:MalloyParser.ExprLiteralContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprCompare.
    def enterExprCompare(self, ctx:MalloyParser.ExprCompareContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprCompare.
    def exitExprCompare(self, ctx:MalloyParser.ExprCompareContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprForRange.
    def enterExprForRange(self, ctx:MalloyParser.ExprForRangeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprForRange.
    def exitExprForRange(self, ctx:MalloyParser.ExprForRangeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprFunc.
    def enterExprFunc(self, ctx:MalloyParser.ExprFuncContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprFunc.
    def exitExprFunc(self, ctx:MalloyParser.ExprFuncContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprAndTree.
    def enterExprAndTree(self, ctx:MalloyParser.ExprAndTreeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprAndTree.
    def exitExprAndTree(self, ctx:MalloyParser.ExprAndTreeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprCast.
    def enterExprCast(self, ctx:MalloyParser.ExprCastContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprCast.
    def exitExprCast(self, ctx:MalloyParser.ExprCastContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprAggFunc.
    def enterExprAggFunc(self, ctx:MalloyParser.ExprAggFuncContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprAggFunc.
    def exitExprAggFunc(self, ctx:MalloyParser.ExprAggFuncContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprTimeTrunc.
    def enterExprTimeTrunc(self, ctx:MalloyParser.ExprTimeTruncContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprTimeTrunc.
    def exitExprTimeTrunc(self, ctx:MalloyParser.ExprTimeTruncContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprAggregate.
    def enterExprAggregate(self, ctx:MalloyParser.ExprAggregateContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprAggregate.
    def exitExprAggregate(self, ctx:MalloyParser.ExprAggregateContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprLogicalAnd.
    def enterExprLogicalAnd(self, ctx:MalloyParser.ExprLogicalAndContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprLogicalAnd.
    def exitExprLogicalAnd(self, ctx:MalloyParser.ExprLogicalAndContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprFieldPath.
    def enterExprFieldPath(self, ctx:MalloyParser.ExprFieldPathContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprFieldPath.
    def exitExprFieldPath(self, ctx:MalloyParser.ExprFieldPathContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprMulDiv.
    def enterExprMulDiv(self, ctx:MalloyParser.ExprMulDivContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprMulDiv.
    def exitExprMulDiv(self, ctx:MalloyParser.ExprMulDivContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprSafeCast.
    def enterExprSafeCast(self, ctx:MalloyParser.ExprSafeCastContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprSafeCast.
    def exitExprSafeCast(self, ctx:MalloyParser.ExprSafeCastContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprOrTree.
    def enterExprOrTree(self, ctx:MalloyParser.ExprOrTreeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprOrTree.
    def exitExprOrTree(self, ctx:MalloyParser.ExprOrTreeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprNot.
    def enterExprNot(self, ctx:MalloyParser.ExprNotContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprNot.
    def exitExprNot(self, ctx:MalloyParser.ExprNotContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprDuration.
    def enterExprDuration(self, ctx:MalloyParser.ExprDurationContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprDuration.
    def exitExprDuration(self, ctx:MalloyParser.ExprDurationContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprApply.
    def enterExprApply(self, ctx:MalloyParser.ExprApplyContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprApply.
    def exitExprApply(self, ctx:MalloyParser.ExprApplyContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprWarnLike.
    def enterExprWarnLike(self, ctx:MalloyParser.ExprWarnLikeContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprWarnLike.
    def exitExprWarnLike(self, ctx:MalloyParser.ExprWarnLikeContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprFieldProps.
    def enterExprFieldProps(self, ctx:MalloyParser.ExprFieldPropsContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprFieldProps.
    def exitExprFieldProps(self, ctx:MalloyParser.ExprFieldPropsContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprCoalesce.
    def enterExprCoalesce(self, ctx:MalloyParser.ExprCoalesceContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprCoalesce.
    def exitExprCoalesce(self, ctx:MalloyParser.ExprCoalesceContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprUngroup.
    def enterExprUngroup(self, ctx:MalloyParser.ExprUngroupContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprUngroup.
    def exitExprUngroup(self, ctx:MalloyParser.ExprUngroupContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprLiteralRecord.
    def enterExprLiteralRecord(self, ctx:MalloyParser.ExprLiteralRecordContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprLiteralRecord.
    def exitExprLiteralRecord(self, ctx:MalloyParser.ExprLiteralRecordContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprWarnIn.
    def enterExprWarnIn(self, ctx:MalloyParser.ExprWarnInContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprWarnIn.
    def exitExprWarnIn(self, ctx:MalloyParser.ExprWarnInContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprPick.
    def enterExprPick(self, ctx:MalloyParser.ExprPickContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprPick.
    def exitExprPick(self, ctx:MalloyParser.ExprPickContext):
        pass


    # Enter a parse tree produced by MalloyParser#exprPathlessAggregate.
    def enterExprPathlessAggregate(self, ctx:MalloyParser.ExprPathlessAggregateContext):
        pass

    # Exit a parse tree produced by MalloyParser#exprPathlessAggregate.
    def exitExprPathlessAggregate(self, ctx:MalloyParser.ExprPathlessAggregateContext):
        pass


    # Enter a parse tree produced by MalloyParser#partialCompare.
    def enterPartialCompare(self, ctx:MalloyParser.PartialCompareContext):
        pass

    # Exit a parse tree produced by MalloyParser#partialCompare.
    def exitPartialCompare(self, ctx:MalloyParser.PartialCompareContext):
        pass


    # Enter a parse tree produced by MalloyParser#partialTest.
    def enterPartialTest(self, ctx:MalloyParser.PartialTestContext):
        pass

    # Exit a parse tree produced by MalloyParser#partialTest.
    def exitPartialTest(self, ctx:MalloyParser.PartialTestContext):
        pass


    # Enter a parse tree produced by MalloyParser#partialAllowedFieldExpr.
    def enterPartialAllowedFieldExpr(self, ctx:MalloyParser.PartialAllowedFieldExprContext):
        pass

    # Exit a parse tree produced by MalloyParser#partialAllowedFieldExpr.
    def exitPartialAllowedFieldExpr(self, ctx:MalloyParser.PartialAllowedFieldExprContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldExprList.
    def enterFieldExprList(self, ctx:MalloyParser.FieldExprListContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldExprList.
    def exitFieldExprList(self, ctx:MalloyParser.FieldExprListContext):
        pass


    # Enter a parse tree produced by MalloyParser#pickStatement.
    def enterPickStatement(self, ctx:MalloyParser.PickStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#pickStatement.
    def exitPickStatement(self, ctx:MalloyParser.PickStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#pick.
    def enterPick(self, ctx:MalloyParser.PickContext):
        pass

    # Exit a parse tree produced by MalloyParser#pick.
    def exitPick(self, ctx:MalloyParser.PickContext):
        pass


    # Enter a parse tree produced by MalloyParser#caseStatement.
    def enterCaseStatement(self, ctx:MalloyParser.CaseStatementContext):
        pass

    # Exit a parse tree produced by MalloyParser#caseStatement.
    def exitCaseStatement(self, ctx:MalloyParser.CaseStatementContext):
        pass


    # Enter a parse tree produced by MalloyParser#caseWhen.
    def enterCaseWhen(self, ctx:MalloyParser.CaseWhenContext):
        pass

    # Exit a parse tree produced by MalloyParser#caseWhen.
    def exitCaseWhen(self, ctx:MalloyParser.CaseWhenContext):
        pass


    # Enter a parse tree produced by MalloyParser#recordKey.
    def enterRecordKey(self, ctx:MalloyParser.RecordKeyContext):
        pass

    # Exit a parse tree produced by MalloyParser#recordKey.
    def exitRecordKey(self, ctx:MalloyParser.RecordKeyContext):
        pass


    # Enter a parse tree produced by MalloyParser#recordRef.
    def enterRecordRef(self, ctx:MalloyParser.RecordRefContext):
        pass

    # Exit a parse tree produced by MalloyParser#recordRef.
    def exitRecordRef(self, ctx:MalloyParser.RecordRefContext):
        pass


    # Enter a parse tree produced by MalloyParser#recordExpr.
    def enterRecordExpr(self, ctx:MalloyParser.RecordExprContext):
        pass

    # Exit a parse tree produced by MalloyParser#recordExpr.
    def exitRecordExpr(self, ctx:MalloyParser.RecordExprContext):
        pass


    # Enter a parse tree produced by MalloyParser#argumentList.
    def enterArgumentList(self, ctx:MalloyParser.ArgumentListContext):
        pass

    # Exit a parse tree produced by MalloyParser#argumentList.
    def exitArgumentList(self, ctx:MalloyParser.ArgumentListContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldNameList.
    def enterFieldNameList(self, ctx:MalloyParser.FieldNameListContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldNameList.
    def exitFieldNameList(self, ctx:MalloyParser.FieldNameListContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldCollection.
    def enterFieldCollection(self, ctx:MalloyParser.FieldCollectionContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldCollection.
    def exitFieldCollection(self, ctx:MalloyParser.FieldCollectionContext):
        pass


    # Enter a parse tree produced by MalloyParser#collectionWildCard.
    def enterCollectionWildCard(self, ctx:MalloyParser.CollectionWildCardContext):
        pass

    # Exit a parse tree produced by MalloyParser#collectionWildCard.
    def exitCollectionWildCard(self, ctx:MalloyParser.CollectionWildCardContext):
        pass


    # Enter a parse tree produced by MalloyParser#starQualified.
    def enterStarQualified(self, ctx:MalloyParser.StarQualifiedContext):
        pass

    # Exit a parse tree produced by MalloyParser#starQualified.
    def exitStarQualified(self, ctx:MalloyParser.StarQualifiedContext):
        pass


    # Enter a parse tree produced by MalloyParser#taggedRef.
    def enterTaggedRef(self, ctx:MalloyParser.TaggedRefContext):
        pass

    # Exit a parse tree produced by MalloyParser#taggedRef.
    def exitTaggedRef(self, ctx:MalloyParser.TaggedRefContext):
        pass


    # Enter a parse tree produced by MalloyParser#refExpr.
    def enterRefExpr(self, ctx:MalloyParser.RefExprContext):
        pass

    # Exit a parse tree produced by MalloyParser#refExpr.
    def exitRefExpr(self, ctx:MalloyParser.RefExprContext):
        pass


    # Enter a parse tree produced by MalloyParser#collectionMember.
    def enterCollectionMember(self, ctx:MalloyParser.CollectionMemberContext):
        pass

    # Exit a parse tree produced by MalloyParser#collectionMember.
    def exitCollectionMember(self, ctx:MalloyParser.CollectionMemberContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldPath.
    def enterFieldPath(self, ctx:MalloyParser.FieldPathContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldPath.
    def exitFieldPath(self, ctx:MalloyParser.FieldPathContext):
        pass


    # Enter a parse tree produced by MalloyParser#joinName.
    def enterJoinName(self, ctx:MalloyParser.JoinNameContext):
        pass

    # Exit a parse tree produced by MalloyParser#joinName.
    def exitJoinName(self, ctx:MalloyParser.JoinNameContext):
        pass


    # Enter a parse tree produced by MalloyParser#fieldName.
    def enterFieldName(self, ctx:MalloyParser.FieldNameContext):
        pass

    # Exit a parse tree produced by MalloyParser#fieldName.
    def exitFieldName(self, ctx:MalloyParser.FieldNameContext):
        pass


    # Enter a parse tree produced by MalloyParser#sqlExploreNameRef.
    def enterSqlExploreNameRef(self, ctx:MalloyParser.SqlExploreNameRefContext):
        pass

    # Exit a parse tree produced by MalloyParser#sqlExploreNameRef.
    def exitSqlExploreNameRef(self, ctx:MalloyParser.SqlExploreNameRefContext):
        pass


    # Enter a parse tree produced by MalloyParser#nameSQLBlock.
    def enterNameSQLBlock(self, ctx:MalloyParser.NameSQLBlockContext):
        pass

    # Exit a parse tree produced by MalloyParser#nameSQLBlock.
    def exitNameSQLBlock(self, ctx:MalloyParser.NameSQLBlockContext):
        pass


    # Enter a parse tree produced by MalloyParser#connectionName.
    def enterConnectionName(self, ctx:MalloyParser.ConnectionNameContext):
        pass

    # Exit a parse tree produced by MalloyParser#connectionName.
    def exitConnectionName(self, ctx:MalloyParser.ConnectionNameContext):
        pass


    # Enter a parse tree produced by MalloyParser#tripFilterString.
    def enterTripFilterString(self, ctx:MalloyParser.TripFilterStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#tripFilterString.
    def exitTripFilterString(self, ctx:MalloyParser.TripFilterStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#tickFilterString.
    def enterTickFilterString(self, ctx:MalloyParser.TickFilterStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#tickFilterString.
    def exitTickFilterString(self, ctx:MalloyParser.TickFilterStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#filterString.
    def enterFilterString(self, ctx:MalloyParser.FilterStringContext):
        pass

    # Exit a parse tree produced by MalloyParser#filterString.
    def exitFilterString(self, ctx:MalloyParser.FilterStringContext):
        pass


    # Enter a parse tree produced by MalloyParser#debugExpr.
    def enterDebugExpr(self, ctx:MalloyParser.DebugExprContext):
        pass

    # Exit a parse tree produced by MalloyParser#debugExpr.
    def exitDebugExpr(self, ctx:MalloyParser.DebugExprContext):
        pass


    # Enter a parse tree produced by MalloyParser#debugPartial.
    def enterDebugPartial(self, ctx:MalloyParser.DebugPartialContext):
        pass

    # Exit a parse tree produced by MalloyParser#debugPartial.
    def exitDebugPartial(self, ctx:MalloyParser.DebugPartialContext):
        pass


    # Enter a parse tree produced by MalloyParser#experimentalStatementForTesting.
    def enterExperimentalStatementForTesting(self, ctx:MalloyParser.ExperimentalStatementForTestingContext):
        pass

    # Exit a parse tree produced by MalloyParser#experimentalStatementForTesting.
    def exitExperimentalStatementForTesting(self, ctx:MalloyParser.ExperimentalStatementForTestingContext):
        pass



del MalloyParser


#!/usr/bin/env python
# /// script
# dependencies = ["sidemantic[serve]", "duckdb"]
# ///
"""Example demonstrating chart generation from semantic layer queries.

Run with: uv run examples/integrations/chart_example.py
"""

from pathlib import Path

from sidemantic import SemanticLayer, load_from_directory
from sidemantic.charts import chart_to_vega, create_chart


def setup_demo_data(layer):
    """Create sample sales data."""
    layer.conn.execute("""
        CREATE TABLE sales (
            sale_id INTEGER,
            sale_date DATE,
            product_category VARCHAR,
            region VARCHAR,
            amount DECIMAL(10, 2)
        )
    """)

    layer.conn.execute("""
        INSERT INTO sales VALUES
            (1, '2024-01-15', 'Electronics', 'North', 1200.00),
            (2, '2024-01-20', 'Electronics', 'South', 800.00),
            (3, '2024-02-05', 'Clothing', 'North', 450.00),
            (4, '2024-02-10', 'Electronics', 'West', 1500.00),
            (5, '2024-02-15', 'Clothing', 'South', 600.00),
            (6, '2024-03-01', 'Electronics', 'North', 900.00),
            (7, '2024-03-10', 'Clothing', 'West', 750.00),
            (8, '2024-03-15', 'Electronics', 'South', 1100.00),
            (9, '2024-04-05', 'Clothing', 'North', 550.00),
            (10, '2024-04-20', 'Electronics', 'West', 1300.00)
    """)


def main():
    # Use the ecommerce example models
    example_dir = Path(__file__).parent.parent / "ecommerce_sql_yml" / "models"

    if not example_dir.exists():
        print(f"Error: Example models not found at {example_dir}")
        print("This example requires the ecommerce_sql_yml models")
        return

    # Load semantic layer
    layer = SemanticLayer()
    load_from_directory(layer, str(example_dir))

    # Setup demo data in the ecommerce DB
    db_path = Path(__file__).parent.parent / "ecommerce_sql_yml" / "data" / "ecommerce.db"
    if not db_path.exists():
        print(f"Error: Database not found. Run: uv run {db_path.parent}/create_db.py")
        return

    # Use the existing database
    layer = SemanticLayer(connection=f"duckdb:///{db_path.absolute()}")
    load_from_directory(layer, str(example_dir))

    print("=" * 80)
    print("Chart Generation Examples")
    print("=" * 80)
    print()

    # Example 1: Revenue trend over time
    print("Example 1: Monthly Revenue Trend (Area Chart)")
    print("-" * 40)

    sql = layer.compile(
        dimensions=["orders.created_at__month"],
        metrics=["orders.revenue"],
        order_by=["orders.created_at__month"],
    )

    result = layer.conn.execute(sql)
    data = [dict(zip([d[0] for d in result.description], row)) for row in result.fetchall()]

    chart = create_chart(
        data=data,
        title="Monthly Revenue Trend",
        chart_type="area",
    )

    vega_spec = chart_to_vega(chart)
    print("Chart type: Area")
    print(f"Data points: {len(data)}")
    print(f"Vega spec keys: {list(vega_spec.keys())}")
    print()

    # Example 2: Revenue by product category
    print("Example 2: Revenue by Category (Bar Chart)")
    print("-" * 40)

    sql = layer.compile(
        dimensions=["products.category"],
        metrics=["order_items.net_revenue"],
        order_by=["order_items.net_revenue desc"],
    )

    result = layer.conn.execute(sql)
    data = [dict(zip([d[0] for d in result.description], row)) for row in result.fetchall()]

    chart = create_chart(
        data=data,
        title="Revenue by Product Category",
        chart_type="bar",
    )

    vega_spec = chart_to_vega(chart)
    print("Chart type: Bar")
    print(f"Data points: {len(data)}")
    print(f"Vega spec keys: {list(vega_spec.keys())}")
    print()

    # Example 3: Multiple metrics over time
    print("Example 3: Revenue & Order Count Trend (Multi-Line)")
    print("-" * 40)

    sql = layer.compile(
        dimensions=["orders.created_at__month"],
        metrics=["orders.revenue", "orders.order_count"],
        order_by=["orders.created_at__month"],
    )

    result = layer.conn.execute(sql)
    data = [dict(zip([d[0] for d in result.description], row)) for row in result.fetchall()]

    chart = create_chart(
        data=data,
        title="Revenue & Orders Over Time",
        chart_type="line",
        width=700,
        height=400,
    )

    vega_spec = chart_to_vega(chart)
    print("Chart type: Line (multiple metrics)")
    print(f"Data points: {len(data)}")
    print("Metrics: revenue, order_count")
    print(f"Vega spec keys: {list(vega_spec.keys())}")
    print()

    print("=" * 80)
    print("✓ All charts generated successfully!")
    print()
    print("Key Features:")
    print("  - Smart chart type selection (area for trends, bar for categories)")
    print("  - Beautiful color palette (modern, accessible)")
    print("  - Clean typography and minimal design")
    print("  - Vega-Lite specs ready for client-side rendering")
    print("  - Can also export to PNG with chart_to_png()")
    print("=" * 80)


if __name__ == "__main__":
    main()



"""Semantic layer main API."""

from __future__ import annotations

from pathlib import Path

from sidemantic.core.metric import Metric
from sidemantic.core.model import Model
from sidemantic.core.semantic_graph import SemanticGraph
from sidemantic.sql.generator import SQLGenerator


class SemanticLayer:
    """Main semantic layer interface.

    Provides a high-level API for defining models and querying data.
    """

    def __init__(
        self,
        connection: str | BaseDatabaseAdapter = "duckdb:///:memory:",  # type: ignore # noqa: F821
        dialect: str | None = None,
        auto_register: bool = True,
        use_preaggregations: bool = False,
        preagg_database: str | None = None,
        preagg_schema: str | None = None,
    ):
        """Initialize semantic layer.

        Args:
            connection: Database connection string or adapter instance (default: in-memory DuckDB)
                Supported URLs:
                - duckdb:///:memory: (default)
                - duckdb:///path/to/db.duckdb
                - duckdb://md:database_name (MotherDuck)
                - postgres://user:pass@host:port/dbname
                - bigquery://project_id/dataset_id
                - snowflake://user:password@account/database/schema
                - clickhouse://user:password@host:port/database
                - databricks://token@server-hostname/http-path
                - spark://host:port/database
                - adbc://driver/uri (e.g., adbc://postgresql/postgresql://localhost/mydb)
            dialect: SQL dialect for query generation (optional, inferred from adapter)
            auto_register: Set as current layer for auto-registration (default: True)
            use_preaggregations: Enable automatic pre-aggregation routing (default: False)
            preagg_database: Optional database name for pre-aggregation tables
            preagg_schema: Optional schema name for pre-aggregation tables
        """
        from sidemantic.db.base import BaseDatabaseAdapter

        self.graph = SemanticGraph()
        self.use_preaggregations = use_preaggregations
        self.preagg_database = preagg_database
        self.preagg_schema = preagg_schema

        # Initialize adapter from connection string or use provided adapter
        if isinstance(connection, BaseDatabaseAdapter):
            self.adapter = connection
            self.dialect = dialect or connection.dialect
            self.connection_string = f"{connection.dialect}://custom"
        elif isinstance(connection, str):
            self.connection_string = connection

            if connection.startswith("duckdb://md:"):
                from sidemantic.db.motherduck import MotherDuckAdapter

                self.adapter = MotherDuckAdapter.from_url(connection)
                self.dialect = dialect or "duckdb"
            elif connection.startswith("duckdb://"):
                from sidemantic.db.duckdb import DuckDBAdapter

                self.adapter = DuckDBAdapter.from_url(connection)
                self.dialect = dialect or "duckdb"
            elif connection.startswith(("postgres://", "postgresql://")):
                from sidemantic.db.postgres import PostgreSQLAdapter

                self.adapter = PostgreSQLAdapter.from_url(connection)
                self.dialect = dialect or "postgres"
            elif connection.startswith("bigquery://"):
                from sidemantic.db.bigquery import BigQueryAdapter

                self.adapter = BigQueryAdapter.from_url(connection)
                self.dialect = dialect or "bigquery"
            elif connection.startswith("snowflake://"):
                from sidemantic.db.snowflake import SnowflakeAdapter

                self.adapter = SnowflakeAdapter.from_url(connection)
                self.dialect = dialect or "snowflake"
            elif connection.startswith("clickhouse://"):
                from sidemantic.db.clickhouse import ClickHouseAdapter

                self.adapter = ClickHouseAdapter.from_url(connection)
                self.dialect = dialect or "clickhouse"
            elif connection.startswith("databricks://"):
                from sidemantic.db.databricks import DatabricksAdapter

                self.adapter = DatabricksAdapter.from_url(connection)
                self.dialect = dialect or "databricks"
            elif connection.startswith("spark://"):
                from sidemantic.db.spark import SparkAdapter

                self.adapter = SparkAdapter.from_url(connection)
                self.dialect = dialect or "spark"
            elif connection.startswith("adbc://"):
                from sidemantic.db.adbc import ADBCAdapter

                self.adapter = ADBCAdapter.from_url(connection)
                self.dialect = dialect or self.adapter.dialect
            else:
                raise ValueError(
                    f"Unsupported connection URL: {connection}. "
                    "Supported: duckdb:///, duckdb://md:, postgres://, bigquery://, snowflake://, clickhouse://, databricks://, spark://, adbc://, or BaseDatabaseAdapter instance"
                )
        else:
            raise TypeError(f"connection must be a string URL or BaseDatabaseAdapter instance, got {type(connection)}")

        # Set as current layer for auto-registration
        if auto_register:
            from .registry import set_current_layer

            set_current_layer(self)

    def __enter__(self):
        """Context manager entry - set as current layer."""
        from .registry import set_current_layer

        set_current_layer(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - clear current layer."""
        from .registry import set_current_layer

        set_current_layer(None)

    @property
    def conn(self):
        """Get raw database connection for backward compatibility."""
        return self.adapter.raw_connection

    @conn.setter
    def conn(self, value):
        """Set raw database connection for backward compatibility.

        This updates the adapter's internal connection.
        Used by tests that directly assign connections.
        """
        # Update the adapter's connection
        self.adapter.conn = value

    def add_model(self, model: Model) -> None:
        """Add a model to the semantic layer.

        Args:
            model: Model to add

        Raises:
            ModelValidationError: If model validation fails
        """
        from sidemantic.validation import ModelValidationError, validate_model

        # Skip registration if an identical model instance is already present.
        #
        # This can happen when models are created while a semantic layer context
        # is active. The model auto-registers itself with the current layer and
        # later user code (or tests) may explicitly call ``add_model`` with the
        # same instance. Treating that call as a no-op keeps ``add_model``
        # idempotent without weakening the duplicate detection performed by the
        # underlying graph (which will still raise when a different definition
        # with the same name is added).
        existing = self.graph.models.get(model.name)
        if existing is not None:
            if existing is model:
                return
            if existing.model_dump() == model.model_dump():
                return

        self._normalize_model_table(model)

        errors = validate_model(model)
        if errors:
            raise ModelValidationError(
                f"Model '{model.name}' validation failed:\n" + "\n".join(f"  - {e}" for e in errors)
            )

        self.graph.add_model(model)

    def _normalize_model_table(self, model: Model) -> None:
        """Normalize model.table for the active dialect when needed."""
        if not model.table or model.sql:
            return

        if self.dialect != "duckdb":
            return

        if not self._is_simple_table_reference(model.table):
            return

        normalized = self._normalize_duckdb_table_reference(model.table)
        if normalized != model.table:
            model.table = normalized

    @staticmethod
    def _is_simple_table_reference(table: str) -> bool:
        """Return True if table looks like a plain table reference (not a subquery)."""
        return not any(ch in table for ch in (" ", "\n", "\t", "(", ")"))

    def _normalize_duckdb_table_reference(self, table: str) -> str:
        """Drop unattached catalog qualifiers for DuckDB."""
        import sqlglot
        from sqlglot import exp

        try:
            parsed = sqlglot.parse_one(table, into=exp.Table, dialect="duckdb")
        except Exception:
            return table

        catalog_expr = parsed.args.get("catalog")
        if not catalog_expr:
            return table

        catalog = getattr(catalog_expr, "name", None) or getattr(catalog_expr, "this", None)
        if not catalog:
            return table

        if catalog in self._duckdb_catalogs():
            return table

        parsed.set("catalog", None)
        return parsed.sql(dialect="duckdb")

    def _duckdb_catalogs(self) -> set[str]:
        """Return attached DuckDB catalog names."""
        try:
            result = self.adapter.execute("PRAGMA database_list")
            rows = result.fetchall()
        except Exception:
            return set()

        catalogs = set()
        for row in rows:
            if len(row) >= 2:
                catalogs.add(row[1])
            elif row:
                catalogs.add(row[0])
        return catalogs

    def add_metric(self, measure: Metric) -> None:
        """Add a measure to the semantic layer.

        Args:
            measure: Metric to add

        Raises:
            MetricValidationError: If measure validation fails
        """
        from sidemantic.validation import MetricValidationError, validate_metric

        existing = self.graph.metrics.get(measure.name)
        if existing is not None:
            if existing is measure:
                return
            if existing.model_dump() == measure.model_dump():
                return

        errors = validate_metric(measure, self.graph)
        if errors:
            raise MetricValidationError(
                f"Measure '{measure.name}' validation failed:\n" + "\n".join(f"  - {e}" for e in errors)
            )

        self.graph.add_metric(measure)

    def query(
        self,
        metrics: list[str] | None = None,
        dimensions: list[str] | None = None,
        filters: list[str] | None = None,
        segments: list[str] | None = None,
        order_by: list[str] | None = None,
        limit: int | None = None,
        ungrouped: bool = False,
        parameters: dict[str, any] | None = None,
        use_preaggregations: bool | None = None,
    ):
        """Execute a query against the semantic layer.

        Args:
            metrics: List of metric references (e.g., ["orders.revenue"])
            dimensions: List of dimension references (e.g., ["orders.status", "orders.order_date__month"])
            filters: List of filter expressions (e.g., ["orders.status = 'completed'"])
            segments: List of segment references (e.g., ["orders.active_users"])
            order_by: List of fields to order by
            limit: Maximum number of rows to return
            ungrouped: If True, return raw rows without aggregation (no GROUP BY)
            parameters: Template parameters for Jinja2 rendering
            use_preaggregations: Override pre-aggregation routing setting for this query

        Returns:
            DuckDB relation object (can convert to DataFrame with .df() or .to_df())
        """
        sql = self.compile(
            metrics=metrics,
            dimensions=dimensions,
            filters=filters,
            segments=segments,
            order_by=order_by,
            limit=limit,
            ungrouped=ungrouped,
            parameters=parameters,
            use_preaggregations=use_preaggregations,
        )

        return self.adapter.execute(sql)

    def compile(
        self,
        metrics: list[str] | None = None,
        dimensions: list[str] | None = None,
        filters: list[str] | None = None,
        segments: list[str] | None = None,
        order_by: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
        dialect: str | None = None,
        ungrouped: bool = False,
        parameters: dict[str, any] | None = None,
        use_preaggregations: bool | None = None,
    ) -> str:
        """Compile a query to SQL without executing.

        Args:
            metrics: List of metric references
            dimensions: List of dimension references
            filters: List of filter expressions
            segments: List of segment references (e.g., ["orders.active_users"])
            order_by: List of fields to order by
            limit: Maximum number of rows to return
            offset: Number of rows to skip
            dialect: SQL dialect override (defaults to layer's dialect)
            ungrouped: If True, return raw rows without aggregation (no GROUP BY)
            use_preaggregations: Override pre-aggregation routing setting for this query

        Returns:
            SQL query string

        Raises:
            QueryValidationError: If query validation fails
        """
        from sidemantic.validation import QueryValidationError, validate_query

        metrics = metrics or []
        dimensions = dimensions or []

        # Validate query
        errors = validate_query(metrics, dimensions, self.graph)
        if errors:
            raise QueryValidationError("Query validation failed:\n" + "\n".join(f"  - {e}" for e in errors))

        # Determine if pre-aggregations should be used
        use_preaggs = use_preaggregations if use_preaggregations is not None else self.use_preaggregations

        generator = SQLGenerator(
            self.graph,
            dialect=dialect or self.dialect,
            preagg_database=self.preagg_database,
            preagg_schema=self.preagg_schema,
        )

        return generator.generate(
            metrics=metrics,
            dimensions=dimensions,
            filters=filters,
            segments=segments,
            order_by=order_by,
            limit=limit,
            offset=offset,
            ungrouped=ungrouped,
            parameters=parameters,
            use_preaggregations=use_preaggs,
        )

    def get_model(self, name: str) -> Model:
        """Get model by name.

        Args:
            name: Model name

        Returns:
            Model instance
        """
        return self.graph.get_model(name)

    def get_metric(self, name: str) -> Metric:
        """Get measure by name.

        Args:
            name: Metric name

        Returns:
            Measure instance
        """
        return self.graph.get_metric(name)

    def list_models(self) -> list[str]:
        """List all model names.

        Returns:
            List of model names
        """
        return list(self.graph.models.keys())

    def list_metrics(self) -> list[str]:
        """List all metric names.

        Returns:
            List of metric names
        """
        return list(self.graph.metrics.keys())

    def get_catalog_metadata(self, schema: str = "public") -> dict:
        """Export semantic layer as Postgres-compatible catalog metadata.

        Returns metadata that can be used to populate information_schema
        and pg_catalog tables, enabling Postgres protocol compatibility.

        Similar to Cube.dev's SQL API, this exposes:
        - Models as tables in information_schema.tables
        - Dimensions and metrics as columns in information_schema.columns
        - Relationships as foreign keys in table_constraints

        Args:
            schema: Schema name to use (default: 'public')

        Returns:
            Dictionary containing:
            - tables: List of table metadata
            - columns: List of column metadata
            - constraints: List of constraint metadata
            - key_column_usage: Foreign key column mappings

        Example:
            >>> catalog = layer.get_catalog_metadata()
            >>> # Use for Postgres wire protocol
            >>> for table in catalog['tables']:
            ...     print(f"{table['table_name']}: {len([c for c in catalog['columns'] if c['table_name'] == table['table_name']])} columns")
        """
        from sidemantic.core.catalog import get_catalog_metadata

        return get_catalog_metadata(self.graph, schema=schema)

    @classmethod
    def from_yaml(cls, path: str | Path, connection: str | None = None) -> SemanticLayer:
        """Load semantic layer from native YAML file.

        Args:
            path: Path to YAML file
            connection: Database connection string (overrides connection in YAML file)

        Returns:
            SemanticLayer instance
        """
        import yaml

        from sidemantic.adapters.sidemantic import SidemanticAdapter, substitute_env_vars

        adapter = SidemanticAdapter()
        graph = adapter.parse(path)

        # If connection not provided as parameter, try to read from YAML file
        # (skip for .sql files which may have multi-document YAML frontmatter)
        path_obj = Path(path)
        if connection is None and path_obj.suffix in (".yml", ".yaml"):
            with open(path) as f:
                content = f.read()
            # Substitute environment variables
            content = substitute_env_vars(content)
            data = yaml.safe_load(content)
            if data and "connection" in data:
                connection = data["connection"]

        # Convert dict-style connection config to URL string
        if isinstance(connection, dict):
            connection = cls._connection_dict_to_url(connection)

        # Create layer with connection (or use default if still None)
        if connection:
            layer = cls(connection=connection)
        else:
            layer = cls()
        layer.graph = graph

        return layer

    @staticmethod
    def _connection_dict_to_url(config: dict) -> str:
        """Convert dict-style connection config to URL string.

        Supports YAML connection configurations like:
            connection:
              type: duckdb
              path: data/warehouse.db

            connection:
              type: postgres
              host: localhost
              port: 5432
              database: mydb
              user: myuser
              password: mypass

            connection:
              type: adbc
              driver: postgresql
              uri: postgresql://localhost/mydb

            connection:
              type: adbc
              driver: snowflake
              account: myaccount
              database: mydb

        Args:
            config: Connection configuration dictionary

        Returns:
            Connection URL string
        """
        from urllib.parse import quote, urlencode

        conn_type = config.get("type", "duckdb").lower()

        if conn_type == "duckdb":
            path = config.get("path", ":memory:")
            if path.startswith("md:"):
                return f"duckdb://{path}"
            return f"duckdb:///{path}"

        elif conn_type in ("postgres", "postgresql"):
            host = config.get("host", "localhost")
            port = config.get("port", 5432)
            database = config.get("database", "postgres")
            user = config.get("user", "")
            password = config.get("password", "")

            if user and password:
                return f"postgres://{quote(user)}:{quote(password)}@{host}:{port}/{database}"
            elif user:
                return f"postgres://{quote(user)}@{host}:{port}/{database}"
            else:
                return f"postgres://{host}:{port}/{database}"

        elif conn_type == "bigquery":
            project = config.get("project")
            dataset = config.get("dataset", "")
            if project:
                return f"bigquery://{project}/{dataset}"
            raise ValueError("BigQuery connection requires 'project' field")

        elif conn_type == "snowflake":
            account = config.get("account")
            user = config.get("user", "")
            password = config.get("password", "")
            database = config.get("database", "")
            schema = config.get("schema", "")

            if not account:
                raise ValueError("Snowflake connection requires 'account' field")

            path = f"/{database}" if database else ""
            if schema:
                path += f"/{schema}"

            if user and password:
                return f"snowflake://{quote(user)}:{quote(password)}@{account}{path}"
            elif user:
                return f"snowflake://{quote(user)}@{account}{path}"
            else:
                return f"snowflake://{account}{path}"

        elif conn_type == "clickhouse":
            host = config.get("host", "localhost")
            port = config.get("port", 8123)
            database = config.get("database", "default")
            user = config.get("user", "")
            password = config.get("password", "")

            if user and password:
                return f"clickhouse://{quote(user)}:{quote(password)}@{host}:{port}/{database}"
            elif user:
                return f"clickhouse://{quote(user)}@{host}:{port}/{database}"
            else:
                return f"clickhouse://{host}:{port}/{database}"

        elif conn_type == "databricks":
            server = config.get("server") or config.get("host")
            http_path = config.get("http_path")
            token = config.get("token", "")

            if not server:
                raise ValueError("Databricks connection requires 'server' or 'host' field")
            if not http_path:
                raise ValueError("Databricks connection requires 'http_path' field")

            return f"databricks://{token}@{server}/{http_path}"

        elif conn_type == "spark":
            host = config.get("host", "localhost")
            port = config.get("port", 10000)
            database = config.get("database", "default")
            return f"spark://{host}:{port}/{database}"

        elif conn_type == "adbc":
            # ADBC connection: driver + optional uri + optional params
            driver = config.get("driver")
            if not driver:
                raise ValueError("ADBC connection requires 'driver' field")

            uri = config.get("uri")

            # Build query params from remaining fields
            params = {k: v for k, v in config.items() if k not in ("type", "driver", "uri")}

            if uri:
                params["uri"] = uri

            if params:
                return f"adbc://{driver}?{urlencode(params)}"
            return f"adbc://{driver}"

        else:
            raise ValueError(
                f"Unknown connection type: {conn_type}. "
                "Supported: duckdb, postgres, bigquery, snowflake, clickhouse, databricks, spark, adbc"
            )

    def sql(self, query: str):
        """Execute a SQL query against the semantic layer.

        Rewrites the SQL to use semantic layer metrics/dimensions and executes it.

        Args:
            query: SQL query like "SELECT revenue, status FROM orders WHERE status = 'completed'"

        Returns:
            DuckDB relation object (can convert to DataFrame with .df() or .to_df())

        Raises:
            ValueError: If SQL cannot be rewritten

        Example:
            >>> layer.sql("SELECT orders.revenue, orders.status FROM orders WHERE orders.status = 'completed'")
        """
        from sidemantic.sql.query_rewriter import QueryRewriter

        rewriter = QueryRewriter(self.graph, dialect=self.dialect)
        rewritten_sql = rewriter.rewrite(query)

        return self.adapter.execute(rewritten_sql)

    def to_yaml(self, path: str | Path) -> None:
        """Export semantic layer to native YAML file.

        Args:
            path: Output file path
        """
        from sidemantic.adapters.sidemantic import SidemanticAdapter

        adapter = SidemanticAdapter()
        adapter.export(self.graph, path)



"""MCP server for Sidemantic semantic layer."""

from decimal import Decimal
from pathlib import Path
from typing import Any, Literal

from mcp.server.fastmcp import FastMCP
from pydantic import BaseModel, Field

from sidemantic.core.semantic_layer import SemanticLayer
from sidemantic.loaders import load_from_directory

# Global semantic layer instance
_layer: SemanticLayer | None = None


class ModelInfo(BaseModel):
    """Basic model information."""

    name: str
    table: str | None = None
    dimensions: list[str]
    metrics: list[str]
    relationships: int


class ModelDetail(BaseModel):
    """Detailed model information."""

    name: str
    table: str | None = None
    dimensions: list[dict[str, Any]]
    metrics: list[dict[str, Any]]
    relationships: list[dict[str, Any]]
    source_format: str | None = None
    source_file: str | None = None


class QueryRequest(BaseModel):
    """Query request parameters."""

    dimensions: list[str] = Field(default_factory=list, description="List of dimensions to include")
    metrics: list[str] = Field(default_factory=list, description="List of metrics to include")
    where: str | None = Field(default=None, description="Optional WHERE clause filter")
    order_by: list[str] = Field(default_factory=list, description="List of columns to order by")
    limit: int | None = Field(default=None, description="Optional row limit")


class QueryResult(BaseModel):
    """Query execution result."""

    sql: str
    rows: list[dict[str, Any]]
    row_count: int


class ChartResult(BaseModel):
    """Chart generation result."""

    sql: str
    vega_spec: dict[str, Any]
    png_base64: str
    row_count: int


def initialize_layer(directory: str, db_path: str | None = None) -> SemanticLayer:
    """Initialize the semantic layer with models from directory."""
    global _layer

    # Create connection string
    connection = None
    if db_path:
        if db_path == ":memory:":
            connection = "duckdb:///:memory:"
        else:
            connection = f"duckdb:///{Path(db_path).absolute()}"

    _layer = SemanticLayer(connection=connection)
    load_from_directory(_layer, directory)
    return _layer


def get_layer() -> SemanticLayer:
    """Get the initialized semantic layer."""
    if _layer is None:
        raise RuntimeError("Semantic layer not initialized. Call initialize_layer first.")
    return _layer


def _convert_to_json_compatible(value: Any) -> Any:
    """Convert value to JSON-compatible type.

    Handles Decimal and other non-JSON-serializable types from database results.
    """
    if isinstance(value, Decimal):
        return float(value)
    return value


def _format_join_condition(model_name: str, rel, models: dict[str, Any]) -> str | None:
    related_name = rel.name
    related_model = models.get(related_name)
    if not related_model:
        return None

    if rel.type == "many_to_one":
        fk = rel.foreign_key or f"{related_name}_id"
        pk = rel.primary_key or related_model.primary_key
        return f"{model_name}.{fk} = {related_name}.{pk}"

    if rel.type in ("one_to_many", "one_to_one"):
        if not rel.foreign_key:
            return None
        pk = models[model_name].primary_key
        return f"{related_name}.{rel.foreign_key} = {model_name}.{pk}"

    if rel.type == "many_to_many":
        if rel.through:
            junction_model = models.get(rel.through)
            if not junction_model:
                return None
            junction_self_fk, junction_related_fk = rel.junction_keys()
            if not junction_self_fk or not junction_related_fk:
                return None
            base_pk = models[model_name].primary_key
            related_pk = rel.primary_key or related_model.primary_key
            return (
                f"{model_name}.{base_pk} = {rel.through}.{junction_self_fk} "
                f"AND {rel.through}.{junction_related_fk} = {related_name}.{related_pk}"
            )
        if rel.foreign_key:
            base_pk = models[model_name].primary_key
            return f"{model_name}.{base_pk} = {related_name}.{rel.foreign_key}"

    return None


# Create MCP server
mcp = FastMCP("sidemantic")


@mcp.tool()
def list_models() -> list[dict[str, Any]]:
    """List all available models in the semantic layer.

    Models are the core building blocks of the semantic layer. Each model represents
    a business entity (e.g., orders, customers, products) and contains:
    - Dimensions: attributes you can group by or filter on
    - Metrics: measures you can aggregate (sum, count, average, etc.)
    - Relationships: connections to other models for automatic joins

    Use this to discover what data is available before constructing queries.

    Returns:
        List of models with basic information including name, table, dimensions,
        metrics, and relationship count.
    """
    layer = get_layer()

    models = []
    for model_name, model in layer.graph.models.items():
        models.append(
            {
                "name": model_name,
                "table": model.table,
                "dimensions": [d.name for d in model.dimensions],
                "metrics": [m.name for m in model.metrics],
                "relationships": len(model.relationships),
            }
        )

    return models


@mcp.tool()
def get_models(model_names: list[str]) -> list[dict[str, Any]]:
    """Get detailed information about one or more models.

    Returns comprehensive details about models including:
    - All dimensions with their types, SQL definitions, and descriptions
    - All metrics with their aggregation types, SQL formulas, filters, and descriptions
    - All relationships showing how models connect for joins
    - Source metadata (original format and file)

    Dimension types:
    - categorical: text/enum values for grouping (e.g., status, region)
    - time: timestamps supporting granularity rollups (e.g., created_at)
    - numeric: numbers that can be used in calculations
    - boolean: true/false flags

    Metric aggregations:
    - sum, avg, min, max: numeric aggregations
    - count, count_distinct: counting aggregations
    - median: statistical median

    Special metric types:
    - ratio: division of two metrics
    - derived: formula combining other metrics
    - cumulative: running totals or rolling windows
    - time_comparison: period-over-period calculations (YoY, MoM, etc.)
    - conversion: funnel conversion rates

    Args:
        model_names: List of model names to retrieve details for

    Returns:
        Detailed information for each requested model including all dimensions,
        metrics, relationships, and source metadata.
    """
    layer = get_layer()

    details = []
    for model_name in model_names:
        if model_name not in layer.graph.models:
            continue

        model = layer.graph.models[model_name]

        # Get dimension details
        dims = []
        for dim in model.dimensions:
            dim_info = {
                "name": dim.name,
                "type": dim.type,
                "sql": dim.sql,
            }
            if dim.description:
                dim_info["description"] = dim.description
            dims.append(dim_info)

        # Get metric details
        metrics = []
        for metric in model.metrics:
            metric_info = {
                "name": metric.name,
                "sql": metric.sql,
            }
            # Include aggregation or type
            if metric.agg:
                metric_info["agg"] = metric.agg
            if metric.type:
                metric_info["type"] = metric.type
            if metric.description:
                metric_info["description"] = metric.description
            if metric.filters:
                metric_info["filters"] = metric.filters
            metrics.append(metric_info)

        # Get relationship details
        rels = []
        for rel in model.relationships:
            rel_info = {
                "name": rel.name,
                "type": rel.type,
            }
            if rel.foreign_key:
                rel_info["foreign_key"] = rel.foreign_key
            if rel.primary_key:
                rel_info["primary_key"] = rel.primary_key
            if rel.through:
                rel_info["through"] = rel.through
            if rel.through_foreign_key:
                rel_info["through_foreign_key"] = rel.through_foreign_key
            if rel.related_foreign_key:
                rel_info["related_foreign_key"] = rel.related_foreign_key
            join_condition = _format_join_condition(model_name, rel, layer.graph.models)
            if join_condition:
                rel_info["join_condition"] = join_condition
            rels.append(rel_info)

        detail = {
            "name": model_name,
            "table": model.table,
            "dimensions": dims,
            "metrics": metrics,
            "relationships": rels,
        }
        if source_format := getattr(model, "_source_format", None):
            detail["source_format"] = source_format
        if source_file := getattr(model, "_source_file", None):
            detail["source_file"] = source_file

        details.append(detail)

    return details


@mcp.tool()
def run_query(
    dimensions: list[str] | None = None,
    metrics: list[str] | None = None,
    where: str | None = None,
    order_by: list[str] | None = None,
    limit: int | None = None,
) -> dict[str, Any]:
    """Run a query against the semantic layer.

    Sidemantic automatically generates SQL from semantic references and handles joins between models.

    Field References:
    - Use model.field_name format (e.g., "orders.customer_name", "orders.total_revenue")
    - Dimensions and metrics are namespaced by their model

    Time Dimensions:
    - Time dimensions support granularity suffixes using double underscore
    - Available granularities: __year, __quarter, __month, __week, __day, __hour
    - Example: "orders.created_at__month" groups by month
    - Use the base dimension name without suffix for raw timestamp

    Automatic Joins:
    - Reference fields from multiple models to trigger automatic joins
    - Joins are inferred from model relationships
    - Example: ["orders.revenue", "customers.region"] automatically joins orders to customers

    Filters:
    - Use model.field_name in WHERE conditions
    - Standard SQL operators: =, !=, <, >, <=, >=, IN, LIKE, BETWEEN
    - Combine with AND/OR
    - Example: "orders.status = 'completed' AND orders.amount > 100"

    Args:
        dimensions: List of dimension references (e.g., ["orders.customer_name", "orders.created_at__month"])
        metrics: List of metric references (e.g., ["orders.total_revenue", "orders.order_count"])
        where: Optional WHERE clause using model.field_name syntax
        order_by: List of fields to order by with optional "asc" or "desc" (e.g., ["orders.total_revenue desc"])
        limit: Optional row limit

    Returns:
        Query result containing generated SQL, result rows, and row count.

    Examples:
        Simple aggregation:
        - dimensions: ["orders.status"]
        - metrics: ["orders.total_revenue"]

        Time series with granularity:
        - dimensions: ["orders.created_at__month"]
        - metrics: ["orders.total_revenue", "orders.order_count"]

        Cross-model query (automatic join):
        - dimensions: ["customers.region", "products.category"]
        - metrics: ["orders.total_revenue"]

        With filters and sorting:
        - dimensions: ["orders.status"]
        - metrics: ["orders.total_revenue"]
        - where: "orders.created_at >= '2024-01-01'"
        - order_by: ["orders.total_revenue desc"]
        - limit: 10
    """
    layer = get_layer()

    # Compile SQL
    sql = layer.compile(
        dimensions=dimensions or [],
        metrics=metrics or [],
        filters=[where] if where else None,
        order_by=order_by,
        limit=limit,
    )

    # Execute query via adapter (works with all database backends)
    result = layer.adapter.execute(sql)

    # Convert to list of dicts with JSON-compatible values
    rows = result.fetchall()
    columns = [desc[0] for desc in result.description]
    row_dicts = [{col: _convert_to_json_compatible(val) for col, val in zip(columns, row)} for row in rows]

    return {
        "sql": sql,
        "rows": row_dicts,
        "row_count": len(row_dicts),
    }


@mcp.tool()
def create_chart(
    dimensions: list[str] | None = None,
    metrics: list[str] | None = None,
    where: str | None = None,
    order_by: list[str] | None = None,
    limit: int | None = None,
    chart_type: Literal["auto", "bar", "line", "area", "scatter", "point"] = "auto",
    title: str | None = None,
    width: int = 600,
    height: int = 400,
) -> dict[str, Any]:
    """Generate a beautiful chart from a semantic layer query.

    This tool combines query execution with intelligent chart generation, producing
    professional, publication-quality visualizations with carefully designed aesthetic defaults.

    Chart Type Selection (when chart_type="auto"):
    - Time dimension + metrics → Line chart (multiple metrics) or Area chart (single metric)
    - Categorical dimension + metrics → Bar chart
    - Two numeric dimensions → Scatter plot
    - Multiple metrics over time → Multi-line chart

    Visual Design:
    - Modern, accessible color palette (not rainbow defaults)
    - Clean typography with Inter font family
    - Minimal gridlines and chartjunk
    - Responsive tooltips showing all relevant data
    - Smart axis formatting (currency, percentages, thousands separators)
    - Professional spacing and proportions

    Query Semantics (same as run_query):
    - Use model.field_name format for all references
    - Time dimensions support granularity: dimension__month, dimension__year, etc.
    - Automatic joins when referencing multiple models
    - Standard SQL operators in WHERE clause

    Args:
        dimensions: List of dimension references (e.g., ["orders.created_at__month", "customers.region"])
        metrics: List of metric references (e.g., ["orders.total_revenue", "orders.order_count"])
        where: Optional WHERE clause filter using model.field_name syntax
        order_by: List of fields to order by with optional "asc" or "desc"
        limit: Optional row limit
        chart_type: Chart type ("auto" for smart selection, or "bar", "line", "area", "scatter", "point")
        title: Chart title (auto-generated if not provided)
        width: Chart width in pixels (default: 600)
        height: Chart height in pixels (default: 400)

    Returns:
        Chart result containing:
        - sql: Generated SQL query
        - vega_spec: Vega-Lite JSON specification (can be rendered client-side)
        - png_base64: Base64-encoded PNG image (ready for display)
        - row_count: Number of data points in the chart

    Examples:
        Revenue trend over time:
        - dimensions: ["orders.created_at__month"]
        - metrics: ["orders.total_revenue"]
        - title: "Monthly Revenue Trend"

        Top products by revenue:
        - dimensions: ["products.name"]
        - metrics: ["orders.total_revenue"]
        - order_by: ["orders.total_revenue desc"]
        - limit: 10
        - chart_type: "bar"

        Revenue by region and status:
        - dimensions: ["customers.region"]
        - metrics: ["orders.total_revenue"]
        - where: "orders.status = 'completed'"
        - chart_type: "bar"

        Multiple metrics over time:
        - dimensions: ["orders.created_at__month"]
        - metrics: ["orders.total_revenue", "orders.order_count"]
        - chart_type: "line"
    """
    from sidemantic.charts import chart_to_base64_png, chart_to_vega
    from sidemantic.charts import create_chart as make_chart

    layer = get_layer()

    # Compile and execute query (same as run_query)
    sql = layer.compile(
        dimensions=dimensions or [],
        metrics=metrics or [],
        filters=[where] if where else None,
        order_by=order_by,
        limit=limit,
    )

    result = layer.adapter.execute(sql)
    rows = result.fetchall()
    columns = [desc[0] for desc in result.description]
    row_dicts = [{col: _convert_to_json_compatible(val) for col, val in zip(columns, row)} for row in rows]

    if not row_dicts:
        raise ValueError("Query returned no data - cannot create chart")

    # Auto-generate title if not provided
    if title is None:
        title = _generate_chart_title(dimensions or [], metrics or [])

    # Create chart with beautiful defaults
    chart = make_chart(
        data=row_dicts,
        chart_type=chart_type,
        title=title,
        width=width,
        height=height,
    )

    # Export to both formats
    vega_spec = chart_to_vega(chart)
    png_base64 = chart_to_base64_png(chart)

    return {
        "sql": sql,
        "vega_spec": vega_spec,
        "png_base64": png_base64,
        "row_count": len(row_dicts),
    }


def _generate_chart_title(dimensions: list[str], metrics: list[str]) -> str:
    """Generate a descriptive title from query parameters."""
    if not metrics:
        return "Data Visualization"

    # Format metric names
    metric_names = [_format_field_name(m) for m in metrics]

    if len(metric_names) == 1:
        title = metric_names[0]
    elif len(metric_names) == 2:
        title = f"{metric_names[0]} & {metric_names[1]}"
    else:
        title = f"{metric_names[0]} & {len(metric_names) - 1} more"

    # Add dimension context if present
    if dimensions:
        dim_name = _format_field_name(dimensions[0])
        title = f"{title} by {dim_name}"

    return title


def _format_field_name(field: str) -> str:
    """Format field name for display (remove model prefix, format as title)."""
    # Remove model prefix
    if "." in field:
        _, field = field.rsplit(".", 1)

    # Handle granularity suffix
    if "__" in field:
        base, granularity = field.rsplit("__", 1)
        field = f"{base} ({granularity})"

    # Convert to title case
    return field.replace("_", " ").title()



"""Malloy adapter for importing/exporting Malloy semantic models.

Uses ANTLR4-generated parser from official Malloy grammar files.
"""

from __future__ import annotations

import re
from pathlib import Path

from antlr4 import CommonTokenStream, InputStream

from sidemantic.adapters.base import BaseAdapter
from sidemantic.adapters.malloy_grammar import MalloyLexer, MalloyParser, MalloyParserVisitor
from sidemantic.core.dimension import Dimension
from sidemantic.core.metric import Metric
from sidemantic.core.model import Model
from sidemantic.core.relationship import Relationship
from sidemantic.core.segment import Segment
from sidemantic.core.semantic_graph import SemanticGraph


class MalloyModelVisitor(MalloyParserVisitor):
    """Visitor that extracts semantic model information from Malloy AST."""

    def __init__(self):
        self.models: list[Model] = []
        # Imports: list of (file_path, items) where items is list of (name, alias) or None for import-all
        self.imports: list[tuple[str, list[tuple[str, str | None]] | None]] = []
        self.current_model_name: str | None = None
        self.current_table: str | None = None
        self.current_sql: str | None = None
        self.current_primary_key: str = "id"
        self.current_description: str | None = None
        self.current_dimensions: list[Dimension] = []
        self.current_metrics: list[Metric] = []
        self.current_relationships: list[Relationship] = []
        self.current_segments: list[Segment] = []

    def _reset_current(self):
        """Reset current model state."""
        self.current_model_name = None
        self.current_table = None
        self.current_sql = None
        self.current_primary_key = "id"
        self.current_description = None
        self.current_dimensions = []
        self.current_metrics = []
        self.current_relationships = []
        self.current_segments = []

    def _parse_annotations(self, tags_ctx) -> str | None:
        """Parse annotations from tags context.

        Malloy annotations:
        - ## Description text -> description (DOC_ANNOTATION)
        - # key: value -> metadata (ANNOTATION)

        We extract:
        - Any ## text as description
        - # desc: value as description

        Returns the description text if found.
        """
        if tags_ctx is None:
            return None

        descriptions = []

        # Iterate through ANNOTATION tokens
        for i in range(tags_ctx.getChildCount()):
            child = tags_ctx.getChild(i)
            if child is not None:
                text = child.getText()

                # ## is a doc annotation (description)
                if text.startswith("##"):
                    # Strip ## and whitespace
                    desc = text[2:].strip()
                    if desc:
                        descriptions.append(desc)
                # # is a tag annotation
                elif text.startswith("#"):
                    # Strip # and check for desc: or description:
                    tag_text = text[1:].strip()
                    # Common patterns: desc: value, description: value
                    if tag_text.lower().startswith("desc:"):
                        desc = tag_text[5:].strip()
                        if desc:
                            descriptions.append(desc)
                    elif tag_text.lower().startswith("description:"):
                        desc = tag_text[12:].strip()
                        if desc:
                            descriptions.append(desc)

        if descriptions:
            return " ".join(descriptions)
        return None

    def visitImportStatement(self, ctx: MalloyParser.ImportStatementContext):  # noqa: N802
        """Visit import statement and extract dependencies.

        Malloy import syntax:
            import 'path/to/file.malloy'                    # Import all sources
            import { source1, source2 } from 'file.malloy'  # Named imports
            import { source1 is alias1 } from 'file.malloy' # Aliased imports
        """
        import_url = ctx.importURL()
        if not import_url:
            return self.visitChildren(ctx)

        # Extract file path from string literal
        url_string = import_url.string()
        file_path = self._extract_string(self._get_text(url_string)) if url_string else ""

        if not file_path:
            return self.visitChildren(ctx)

        # Check for selective imports: import { x, y } from 'file'
        import_select = ctx.importSelect()
        if import_select:
            items = []
            for import_item in import_select.importItem():
                # Grammar uses id_() method (underscore to avoid Python keyword)
                ids = import_item.id_()
                if ids:
                    # First id is the source name, second (if present) is the alias
                    name = self._get_text(ids[0])
                    alias = self._get_text(ids[1]) if len(ids) > 1 else None
                    items.append((name, alias))
            self.imports.append((file_path, items))
        else:
            # Import all sources from file
            self.imports.append((file_path, None))

        return self.visitChildren(ctx)

    def _get_text(self, ctx) -> str:
        """Get text from context, preserving whitespace.

        ANTLR's getText() concatenates tokens without whitespace.
        We reconstruct the original text using the token stream.
        """
        if ctx is None:
            return ""

        # Try to get original text with whitespace from token stream
        try:
            start = ctx.start
            stop = ctx.stop
            if start and stop:
                input_stream = start.getInputStream()
                return input_stream.getText(start.start, stop.stop)
        except (AttributeError, TypeError):
            pass

        # Fallback to getText()
        return ctx.getText()

    def _extract_string(self, text: str) -> str:
        """Remove quotes from string literal."""
        if not text:
            return text
        # Remove single, double, or backtick quotes
        if (
            (text.startswith("'") and text.endswith("'"))
            or (text.startswith('"') and text.endswith('"'))
            or (text.startswith("`") and text.endswith("`"))
        ):
            return text[1:-1]
        # Handle triple quotes
        if text.startswith("'''") and text.endswith("'''"):
            return text[3:-3]
        if text.startswith('"""') and text.endswith('"""'):
            return text[3:-3]
        return text

    def _infer_dimension_type(self, sql: str, name: str) -> str:
        """Infer dimension type from SQL expression and name."""
        sql_lower = sql.lower() if sql else ""
        name_lower = name.lower()

        # Time dimension detection
        time_patterns = [
            "date_trunc",
            "::date",
            "::timestamp",
            "::timestamptz",
            "extract",
            "strftime",
            "to_date",
            "to_timestamp",
        ]
        if any(p in sql_lower for p in time_patterns):
            return "time"

        time_name_patterns = ["date", "time", "timestamp", "_at", "created", "updated"]
        if any(p in name_lower for p in time_name_patterns):
            return "time"

        # Boolean detection - comparison that yields true/false
        if re.search(r"[<>=!]+\s*\S", sql):
            # Check if it's a simple comparison (boolean result)
            # But not if it's part of a CASE/pick statement
            if "pick" not in sql_lower and "case" not in sql_lower:
                return "boolean"

        # Numeric detection
        if re.search(r"[+\-*/]", sql) and "||" not in sql:
            return "numeric"

        return "categorical"

    def _extract_granularity(self, sql: str) -> str | None:
        """Extract time granularity from SQL expression."""
        if not sql:
            return None

        sql_lower = sql.lower()
        valid_granularities = ("second", "minute", "hour", "day", "week", "month", "quarter", "year")

        # DATE_TRUNC('month', field) -> 'month'
        match = re.search(r"date_trunc\s*\(\s*['\"](\w+)['\"]", sql_lower)
        if match:
            granularity = match.group(1)
            if granularity in valid_granularities:
                return granularity

        # .second, .minute, .day, .month, .year etc (Malloy time truncation)
        match = re.search(r"\.(\w+)$", sql_lower)
        if match:
            granularity = match.group(1)
            if granularity in valid_granularities:
                return granularity

        # ::date cast
        if "::date" in sql_lower:
            return "day"

        return None

    def _parse_aggregation(self, expr: str) -> tuple[str | None, str | None]:
        """Parse aggregation function from expression.

        Returns (agg_type, sql_expr) tuple.
        """
        if not expr:
            return None, None

        # Match: count(), sum(x), avg(x), etc.
        match = re.match(r"(\w+)\s*\(\s*(.*?)\s*\)$", expr.strip(), re.DOTALL)
        if match:
            agg_func = match.group(1).lower()
            agg_arg = match.group(2).strip()

            if agg_func in ("count", "sum", "avg", "min", "max"):
                return agg_func, agg_arg if agg_arg else None
            elif agg_func == "count_distinct":
                return "count_distinct", agg_arg

        return None, expr

    def _transform_pick_to_case(self, expr: str) -> str:
        """Transform Malloy pick/when/else to SQL CASE expression."""
        # Pattern: pick 'value' when condition
        # Becomes: CASE WHEN condition THEN 'value' ... END

        lines = expr.strip().split("\n")
        cases = []
        else_value = None

        for line in lines:
            line = line.strip()
            if not line:
                continue

            # Match: pick 'value' when condition
            pick_match = re.match(r"pick\s+(.+?)\s+when\s+(.+)$", line, re.IGNORECASE)
            if pick_match:
                value = pick_match.group(1).strip()
                condition = pick_match.group(2).strip()
                cases.append(f"WHEN {condition} THEN {value}")
                continue

            # Match: else 'value'
            else_match = re.match(r"else\s+(.+)$", line, re.IGNORECASE)
            if else_match:
                else_value = else_match.group(1).strip()

        if cases:
            case_str = "CASE " + " ".join(cases)
            if else_value:
                case_str += f" ELSE {else_value}"
            case_str += " END"
            return case_str

        return expr

    def visitDefineSourceStatement(self, ctx: MalloyParser.DefineSourceStatementContext):  # noqa: N802
        """Visit source: name is ... statement."""
        # Get statement-level tags (before 'source:' keyword)
        # These apply to all sources in the statement if there's only one,
        # or can be overridden by source-specific tags
        stmt_tags = ctx.tags()
        stmt_description = self._parse_annotations(stmt_tags) if stmt_tags else None

        # Get source definitions
        source_list = ctx.sourcePropertyList()
        if source_list:
            source_defs = source_list.sourceDefinition()
            for source_def in source_defs:
                self._reset_current()
                self._process_source_definition(source_def)

                # If source has no description but statement does, use statement description
                # (only for single-source statements, or as fallback)
                if self.current_description is None and stmt_description is not None:
                    self.current_description = stmt_description

                if self.current_model_name:
                    model = Model(
                        name=self.current_model_name,
                        table=self.current_table,
                        sql=self.current_sql,
                        primary_key=self.current_primary_key,
                        description=self.current_description,
                        dimensions=self.current_dimensions,
                        metrics=self.current_metrics,
                        relationships=self.current_relationships,
                        segments=self.current_segments,
                    )
                    self.models.append(model)

        return self.visitChildren(ctx)

    def _process_source_definition(self, ctx: MalloyParser.SourceDefinitionContext):
        """Process a single source definition."""
        # Get annotations from tags
        tags = ctx.tags()
        if tags:
            self.current_description = self._parse_annotations(tags)

        # Get source name
        name_def = ctx.sourceNameDef()
        if name_def:
            self.current_model_name = self._get_text(name_def)

        # Process the source expression (sqExplore -> sqExpr)
        sq_explore = ctx.sqExplore()
        if sq_explore:
            sq_expr = sq_explore.sqExpr()
            self._process_sq_expr(sq_expr)

    def _process_sq_expr(self, ctx: MalloyParser.SqExprContext):
        """Process source expression - table, sql, or extended source."""
        if ctx is None:
            return

        # Check for table reference: connection.table('path')
        if isinstance(ctx, MalloyParser.SQTableContext):
            explore_table = ctx.exploreTable()
            if explore_table:
                table_path = explore_table.tablePath()
                if table_path:
                    self.current_table = self._extract_string(self._get_text(table_path))
            return

        # Check for SQL reference: connection.sql('...')
        if isinstance(ctx, MalloyParser.SQSQLContext):
            sql_source = ctx.sqlSource()
            if sql_source:
                # Extract SQL string
                sql_string = sql_source.sqlString()
                if sql_string:
                    # sqlString includes triple quotes, extract the content
                    self.current_sql = self._extract_string(self._get_text(sql_string))
                else:
                    short_string = sql_source.shortString()
                    if short_string:
                        self.current_sql = self._extract_string(self._get_text(short_string))
            return

        # Check for extended source: base extend { ... }
        if isinstance(ctx, MalloyParser.SQExtendedSourceContext):
            # First process the base source
            base_sq_expr = ctx.sqExpr()
            if base_sq_expr:
                self._process_sq_expr(base_sq_expr)

            # Then process the extend block
            explore_props = ctx.exploreProperties()
            if explore_props:
                self._process_explore_properties(explore_props)
            return

        # Check for source with include: base include { ... }
        if isinstance(ctx, MalloyParser.SQIncludeContext):
            base_sq_expr = ctx.sqExpr()
            if base_sq_expr:
                self._process_sq_expr(base_sq_expr)
            return

        # Check for ID reference (another source name)
        if isinstance(ctx, MalloyParser.SQIDContext):
            # This is a reference to another source - we might need to track extends
            return

        # Check for parenthesized expression
        if isinstance(ctx, MalloyParser.SQParensContext):
            inner = ctx.sqExpr()
            if inner:
                self._process_sq_expr(inner)
            return

    def _process_explore_properties(self, ctx: MalloyParser.ExplorePropertiesContext):
        """Process the extend { ... } block of a source."""
        for stmt in ctx.exploreStatement():
            self._process_explore_statement(stmt)

    def _process_explore_statement(self, ctx: MalloyParser.ExploreStatementContext):
        """Process a single statement in explore properties."""
        # Primary key
        if isinstance(ctx, MalloyParser.DefExplorePrimaryKeyContext):
            field_name = ctx.fieldName()
            if field_name:
                self.current_primary_key = self._get_text(field_name)
            return

        # Dimensions
        if isinstance(ctx, MalloyParser.DefExploreDimension_stubContext):
            def_dims = ctx.defDimensions()
            if def_dims:
                self._process_def_dimensions(def_dims)
            return

        # Measures
        if isinstance(ctx, MalloyParser.DefExploreMeasure_stubContext):
            def_measures = ctx.defMeasures()
            if def_measures:
                self._process_def_measures(def_measures)
            return

        # Joins
        if isinstance(ctx, MalloyParser.DefJoin_stubContext):
            join_stmt = ctx.joinStatement()
            if join_stmt:
                self._process_join_statement(join_stmt)
            return

        # Where (source-level filter -> segment)
        if isinstance(ctx, MalloyParser.DefExploreWhere_stubContext):
            where_stmt = ctx.whereStatement()
            if where_stmt:
                self._process_where_as_segment(where_stmt)
            return

    def _process_def_dimensions(self, ctx: MalloyParser.DefDimensionsContext):
        """Process dimension: statements."""
        def_list = ctx.defList()
        if not def_list:
            return

        for field_def in def_list.fieldDef():
            self._process_dimension_def(field_def)

    def _process_dimension_def(self, ctx: MalloyParser.FieldDefContext):
        """Process a single dimension definition."""
        name_def = ctx.fieldNameDef()
        if not name_def:
            return

        name = self._get_text(name_def)

        # Get annotations from tags
        tags = ctx.tags()
        description = self._parse_annotations(tags) if tags else None

        # Get the expression
        field_expr = ctx.fieldExpr()
        sql = self._get_text(field_expr) if field_expr else name

        # Transform pick/when to CASE
        if "pick" in sql.lower():
            sql = self._transform_pick_to_case(sql)

        # Infer type
        dim_type = self._infer_dimension_type(sql, name)

        # Extract granularity for time dimensions
        granularity = None
        if dim_type == "time":
            granularity = self._extract_granularity(sql)

        self.current_dimensions.append(
            Dimension(
                name=name,
                type=dim_type,
                sql=sql,
                granularity=granularity,
                description=description,
            )
        )

    def _process_def_measures(self, ctx: MalloyParser.DefMeasuresContext):
        """Process measure: statements."""
        def_list = ctx.defList()
        if not def_list:
            return

        for field_def in def_list.fieldDef():
            self._process_measure_def(field_def)

    def _process_measure_def(self, ctx: MalloyParser.FieldDefContext):
        """Process a single measure definition."""
        name_def = ctx.fieldNameDef()
        if not name_def:
            return

        name = self._get_text(name_def)

        # Get annotations from tags
        tags = ctx.tags()
        description = self._parse_annotations(tags) if tags else None

        # Get the expression
        field_expr = ctx.fieldExpr()
        expr_text = self._get_text(field_expr) if field_expr else ""

        # Check for filtered measure: count() { where: ... }
        filters = None
        if isinstance(field_expr, MalloyParser.ExprFieldPropsContext):
            # Has field properties (like { where: ... })
            inner_expr = field_expr.fieldExpr()
            props = field_expr.fieldProperties()

            expr_text = self._get_text(inner_expr) if inner_expr else ""

            if props:
                for prop_stmt in props.fieldPropertyStatement():
                    if isinstance(prop_stmt, MalloyParser.WhereStatementContext) or hasattr(
                        prop_stmt, "whereStatement"
                    ):
                        where_stmt = (
                            prop_stmt
                            if isinstance(prop_stmt, MalloyParser.WhereStatementContext)
                            else getattr(prop_stmt, "whereStatement", lambda: None)()
                        )
                        if where_stmt:
                            filter_list = where_stmt.filterClauseList()
                            if filter_list:
                                filters = [self._get_text(f) for f in filter_list.fieldExpr()]

        # Parse aggregation
        agg, sql = self._parse_aggregation(expr_text)

        # Determine metric type
        metric_type = None
        if agg is None and sql:
            # All non-aggregation expressions are derived metrics
            # Ratio type requires numerator/denominator metric references which we can't
            # reliably extract from arbitrary Malloy expressions
            metric_type = "derived"

        self.current_metrics.append(
            Metric(
                name=name,
                type=metric_type,
                agg=agg,
                sql=sql,
                filters=filters,
                description=description,
            )
        )

    def _process_join_statement(self, ctx: MalloyParser.JoinStatementContext):
        """Process join_one/join_many statements."""
        # Determine join type
        if isinstance(ctx, MalloyParser.DefJoinOneContext):
            rel_type = "many_to_one"
            join_list = ctx.joinList()
        elif isinstance(ctx, MalloyParser.DefJoinManyContext):
            rel_type = "one_to_many"
            join_list = ctx.joinList()
        elif isinstance(ctx, MalloyParser.DefJoinCrossContext):
            rel_type = "one_to_one"  # Cross joins mapped to one_to_one
            join_list = ctx.joinList()
        else:
            return

        if not join_list:
            return

        for join_def in join_list.joinDef():
            self._process_join_def(join_def, rel_type)

    def _process_join_def(self, ctx: MalloyParser.JoinDefContext, rel_type: str):
        """Process a single join definition."""
        join_from = ctx.joinFrom()
        if not join_from:
            return

        # Get join name (could be alias is source or just source)
        join_name_def = join_from.joinNameDef()
        name = self._get_text(join_name_def) if join_name_def else None

        if not name:
            return

        # Check if there's an isExplore (alias is source)
        # Note: We could use this to track the target model for extended support
        # is_explore = join_from.isExplore()
        # if is_explore:
        #     sq_expr = is_explore.sqExpr()
        #     target_model = self._get_text(sq_expr) if sq_expr else name

        # Get foreign key from 'with' clause
        foreign_key = None
        if isinstance(ctx, MalloyParser.JoinWithContext):
            field_expr = ctx.fieldExpr()
            if field_expr:
                foreign_key = self._get_text(field_expr)
        elif isinstance(ctx, MalloyParser.JoinOnContext):
            # Parse 'on' clause to extract FK if possible
            join_expr = ctx.joinExpression()
            if join_expr:
                # Try to extract FK from simple equality: fk = other.pk
                expr_text = self._get_text(join_expr)
                # Simple heuristic: first identifier before = is often the FK
                match = re.match(r"(\w+)\s*=", expr_text)
                if match:
                    foreign_key = match.group(1)

        self.current_relationships.append(
            Relationship(
                name=name,
                type=rel_type,
                foreign_key=foreign_key,
            )
        )

    def _process_where_as_segment(self, ctx: MalloyParser.WhereStatementContext):
        """Process source-level where clause as a segment."""
        filter_list = ctx.filterClauseList()
        if not filter_list:
            return

        for i, filter_expr in enumerate(filter_list.fieldExpr()):
            sql = self._get_text(filter_expr)
            self.current_segments.append(
                Segment(
                    name=f"default_filter_{i}" if i > 0 else "default_filter",
                    sql=sql,
                )
            )


class MalloyAdapter(BaseAdapter):
    """Adapter for importing/exporting Malloy semantic models.

    Transforms Malloy definitions into Sidemantic format:
    - Sources -> Models
    - Dimensions -> Dimensions
    - Measures -> Metrics
    - join_one/join_many -> Relationships
    - Source-level where -> Segments

    Note: Views and queries are skipped as they are not part of
    the semantic model definition.
    """

    def parse(self, source: str | Path) -> SemanticGraph:
        """Parse Malloy files into semantic graph.

        Handles imports by recursively parsing imported files first (depth-first).
        Detects and prevents circular imports.

        Args:
            source: Path to .malloy file or directory

        Returns:
            Semantic graph with imported models
        """
        graph = SemanticGraph()
        source_path = Path(source)

        if source_path.is_dir():
            # Parse all .malloy files in directory
            # When parsing a directory, each file is treated as a "root" file,
            # meaning all of its models are added (no import filter).
            # We use a separate parsed_files set per root file to handle imports,
            # but we track which models have been added to avoid duplicates.
            for malloy_file in source_path.rglob("*.malloy"):
                parsed_files: set[Path] = set()
                self._parse_file(malloy_file, graph, parsed_files, import_filter=None)
        else:
            parsed_files: set[Path] = set()
            self._parse_file(source_path, graph, parsed_files)

        return graph

    def _parse_file(
        self,
        file_path: Path,
        graph: SemanticGraph,
        parsed_files: set[Path],
        import_filter: list[tuple[str, str | None]] | None = None,
    ) -> None:
        """Parse a single Malloy file with import resolution.

        Args:
            file_path: Path to .malloy file
            graph: Semantic graph to add models to
            parsed_files: Set of already-parsed file paths (for cycle detection)
            import_filter: If set, only import these sources (name, alias) pairs.
                          None means import all sources from the file.
        """
        resolved_path = file_path.resolve()

        # Cycle detection: skip if already parsed
        if resolved_path in parsed_files:
            return

        parsed_files.add(resolved_path)

        if not file_path.exists():
            return

        with open(file_path) as f:
            content = f.read()

        # Create ANTLR input stream
        input_stream = InputStream(content)
        lexer = MalloyLexer(input_stream)
        token_stream = CommonTokenStream(lexer)
        parser = MalloyParser(token_stream)

        # Parse the document
        tree = parser.malloyDocument()

        # Visit the tree to extract models and imports
        visitor = MalloyModelVisitor()
        visitor.visit(tree)

        # Process imports first (depth-first) so imported sources are available
        for import_path, import_items in visitor.imports:
            # Resolve import path relative to current file
            import_file = (file_path.parent / import_path).resolve()
            if import_file.exists():
                self._parse_file(import_file, graph, parsed_files, import_items)

        # Add models to graph (with optional filtering for selective imports)
        for model in visitor.models:
            if import_filter is not None:
                # Check if this model should be imported
                matching_import = None
                for name, alias in import_filter:
                    if model.name == name:
                        matching_import = (name, alias)
                        break

                if matching_import is None:
                    continue  # Skip this model, not in import list

                # Apply alias if specified
                name, alias = matching_import
                if alias:
                    model = Model(
                        name=alias,
                        table=model.table,
                        sql=model.sql,
                        description=model.description,
                        extends=model.extends,
                        relationships=model.relationships,
                        primary_key=model.primary_key,
                        dimensions=model.dimensions,
                        metrics=model.metrics,
                        segments=model.segments,
                        pre_aggregations=model.pre_aggregations,
                    )

            # Add to graph if not already present
            if model.name not in graph.models:
                graph.add_model(model)

    def export(self, graph: SemanticGraph, output_path: str | Path) -> None:
        """Export semantic graph to Malloy format.

        Args:
            graph: Semantic graph to export
            output_path: Path to output .malloy file
        """
        output_path = Path(output_path)

        # Resolve inheritance first
        from sidemantic.core.inheritance import resolve_model_inheritance

        resolved_models = resolve_model_inheritance(graph.models)

        # Generate Malloy content
        lines = []
        for model in resolved_models.values():
            source_lines = self._export_source(model)
            lines.extend(source_lines)
            lines.append("")  # Empty line between sources

        output_path.parent.mkdir(parents=True, exist_ok=True)

        with open(output_path, "w") as f:
            f.write("\n".join(lines))

    def _strip_model_prefix(self, sql: str) -> str:
        """Strip {model}. prefix from SQL expression.

        Malloy doesn't need table qualifiers for column references within a source,
        so we remove the {model}. placeholder that sidemantic uses internally.

        Args:
            sql: SQL expression that may contain {model}. prefixes

        Returns:
            SQL expression with {model}. prefixes removed
        """
        return sql.replace("{model}.", "")

    def _export_source(self, model: Model) -> list[str]:
        """Export a model to Malloy source definition.

        Args:
            model: Model to export

        Returns:
            List of lines for the source definition
        """
        lines = []

        # Model description as annotation
        if model.description:
            lines.append(f"# desc: {model.description}")

        # Source header
        if model.table:
            lines.append(f"source: {model.name} is duckdb.table('{model.table}') extend {{")
        elif model.sql:
            # For SQL-based sources
            lines.append(f'source: {model.name} is duckdb.sql("""{model.sql}""") extend {{')
        else:
            lines.append(f"source: {model.name} extend {{")

        # Primary key
        if model.primary_key and model.primary_key != "id":
            lines.append(f"  primary_key: {model.primary_key}")

        # Dimensions
        # Skip dimensions that match the primary key (Malloy auto-exposes the PK column).
        # Skip passthrough dimensions (sql == name) since Malloy auto-exposes underlying
        # table columns. Only export dimensions with actual transformations.
        dims_to_export: list[tuple[Dimension, str]] = []
        for dim in model.dimensions:
            if dim.name == model.primary_key:
                continue
            sql = self._strip_model_prefix(dim.sql or dim.name).strip()
            # Skip passthrough dimensions - Malloy auto-exposes table columns
            if sql == dim.name:
                continue
            dims_to_export.append((dim, sql))

        if dims_to_export:
            lines.append("")
            lines.append("  dimension:")
            for dim, sql in dims_to_export:
                if dim.description:
                    lines.append(f"    # desc: {dim.description}")
                # For time dimensions with granularity, use Malloy's time truncation syntax
                # But only if the SQL doesn't already contain a truncation function
                if dim.type == "time" and dim.granularity:
                    sql_lower = sql.lower()
                    already_has_truncation = (
                        "date_trunc" in sql_lower or "::date" in sql_lower or sql_lower.endswith(f".{dim.granularity}")
                    )
                    if not already_has_truncation:
                        # Append Malloy time accessor: .second, .minute, .hour, .day, .week, .month, .quarter, .year
                        lines.append(f"    {dim.name} is {sql}.{dim.granularity}")
                    else:
                        lines.append(f"    {dim.name} is {sql}")
                else:
                    lines.append(f"    {dim.name} is {sql}")

        # Measures
        if model.metrics:
            lines.append("")
            lines.append("  measure:")
            for metric in model.metrics:
                if metric.description:
                    lines.append(f"    # desc: {metric.description}")
                measure_expr = self._format_measure(metric)
                lines.append(f"    {metric.name} is {measure_expr}")

        # Joins
        for rel in model.relationships:
            lines.append("")
            join_type = "join_one" if rel.type == "many_to_one" else "join_many"
            if rel.foreign_key:
                lines.append(f"  {join_type}: {rel.name} with {rel.foreign_key}")
            else:
                lines.append(f"  {join_type}: {rel.name}")

        lines.append("}")

        return lines

    def _format_measure(self, metric: Metric) -> str:
        """Format a metric as Malloy measure expression.

        Args:
            metric: Metric to format

        Returns:
            Malloy measure expression string
        """
        # Simple aggregation
        if metric.agg:
            if metric.sql:
                sql = self._strip_model_prefix(metric.sql)
                expr = f"{metric.agg}({sql})"
            else:
                expr = f"{metric.agg}()"

            # Add filter if present
            if metric.filters:
                filters = [self._strip_model_prefix(f) for f in metric.filters]
                filter_str = ", ".join(filters)
                expr = f"{expr} {{ where: {filter_str} }}"

            return expr

        # Derived/ratio metric
        if metric.type == "ratio" and metric.numerator and metric.denominator:
            return f"{metric.numerator} / {metric.denominator}"

        # Fallback to sql
        if metric.sql:
            return self._strip_model_prefix(metric.sql)

        return "count()"
